#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.util.scriptlib.help_formatters`"""
from nose.tools import assert_equal, assert_not_equal, assert_true
from nose.plugins.attrib import attr
from plastid.util.scriptlib.help_formatters import shorten_help,\
                                                           format_module_docstring,\
                                                           pyrst_pattern,\
                                                           subst_pattern,\
                                                           link_pattern,\
                                                           _separator


@attr(test="unit")
class TestHelpFormatters():
    def test_pyrst_pattern_matches(self):
        tests = [
            (
                ":py:class:`plastid.genomics.genome_array.GenomeArray`", "py", "class",
                "plastid.genomics.genome_array.GenomeArray", None
            ),
            (
                ":py:mod:`plastid.genomics.genome_array`", "py", "mod",
                "plastid.genomics.genome_array", None
            ),
            (":py:obj:`some_obj`", "py", "obj", "some_obj", None),
            (
                ":class:`plastid.genomics.genome_array.GenomeArray`", None, "class",
                "plastid.genomics.genome_array.GenomeArray", None
            ),
            (
                ":mod:`plastid.genomics.genome_array`", None, "mod",
                "plastid.genomics.genome_array", None
            ),
            (":obj:`some_obj`", None, "obj", "some_obj", None),
            (":obj:`some_obj <with pointer>`", None, "obj", "some_obj", "with pointer"),
            (
                ":domain:role:`some_arg <with pointer>`", "domain", "role", "some_arg",
                "with pointer"
            ),
        ]
        for test, domain, role, argument, pointer in tests:
            groupdict = pyrst_pattern.search(test).groupdict()
            yield self.check_pyrst_pattern, domain, groupdict["domain"]
            yield self.check_pyrst_pattern, role, groupdict["role"]
            yield self.check_pyrst_pattern, argument, groupdict["argument"]
            yield self.check_pyrst_pattern, pointer, groupdict["pointer"]

    @staticmethod
    def check_pyrst_pattern(expected, found, message=""):
        if len(message) == 0:
            message = "Expected %s, found %s." % (expected, found)
        assert_equal(expected, found, message)

    def test_pyrst_pattern_nonmatches(self):
        tests = [
            (":py:"),
            (":py:class:"),
            ("`genome_array`"),
            # missing one or more backticks
            (":py:class:`plastid.genomics.genome_array.GenomeArray"),
            (":py:class:plastid.genomics.genome_array.GenomeArray`"),
            (":py:class:plastid.genomics.genome_array.GenomeArray"),
            # malformed starting token
            ("py:class:`plastid.genomics.genome_array.GenomeArray`"),
        ]
        for test in tests:
            yield self.check_pyrst_pattern, pyrst_pattern.search(
                test
            ), None, "rst pattern matches %s" % test

    def test_subst_pattern_matches(self):
        assert_equal(subst_pattern.search("|test_replacement|").groups()[0], "test_replacement")

    def test_subst_pattern_nonmatches(self):
        tests = ["|bad_test", "bad_test|"]
        for test in tests:
            assert_true(
                subst_pattern.search(test) is None, "substitution pattern matches %s" % test
            )

    def test_link_pattern_matches(self):
        assert_equal(link_pattern.search("`My link`_").groups()[0], "My link")
        assert_equal(link_pattern.search("`Mylink`_").groups()[0], "Mylink")
        assert_equal(
            link_pattern.search("`Mylink (with parentheses)`_").groups()[0],
            "Mylink (with parentheses)"
        )
        assert_equal(link_pattern.search("`Mylink <with_subs>`_").groups()[0], "Mylink")

    def test_link_pattern_nonmatches(self):
        tests = ["|bad_test", "bad_test|"]
        for test in tests:
            assert_true(link_pattern.search(test) is None, "substitution pattern matches %s" % test)

    def test_shorten_help(self):
        assert_not_equal, raw_help1, shortened_help1
        assert_equal(shorten_help(raw_help1), shortened_help1)

    def test_format_module_docstring(self):
        expected = _separator + "\n" + shortened_help1 + "\n" + _separator
        assert_equal(format_module_docstring(raw_help1), expected)
        assert_equal(format_module_docstring(shortened_help1), expected)


#===============================================================================
# INDEX: test data
#===============================================================================

raw_help1 = """Implements array-like data structures called |GenomeArray|, that map numerical values 
to genomic positions (nucleotides) under various configurable mapping rules.
Several different implementations are provided, depending on how the alignment
and/or count data is stored:

Important classes
-----------------
    
    |AbstractGenomeArray|
        Base class for all genome arrays, array-like data structures mapping
        read counts to specific nucleotide positions.  
        
        Defines interfaces for:
            - retrieving vectors (as :py:class:`numpy.ndarray` s) of read counts
              at each position of a |GenomicSegment|
              
            - getting the sum of read counts in a dataset
            
            - toggling normalization of fetched counts to reads-per-million
    
    |MutableAbstractGenomeArray|
        Base class for |GenomeArray| and |SparseGenomeArray|. Contains all interfaces
        from |AbstractGenomeArray|, and additionally defines interfaces for:
        
            - setting values, manually or mathematically, over regions of the
              genome, or over the entire genome, element wise
            
See also
--------
:py:meth:`some_method`
    A method that is useful

:py:class:`plastid.genomics.genome_array.GenomeArray`
    A useful class
            
Parameters
----------
foo1 : footype
    Some parameter

foo2 : footype
    another parameter

Raises
------
FooException
    If the foo doesn't bar
"""

shortened_help1 = """Implements array-like data structures called GenomeArray, that map numerical values 
to genomic positions (nucleotides) under various configurable mapping rules.
Several different implementations are provided, depending on how the alignment
and/or count data is stored:

Important classes
-----------------
    
    AbstractGenomeArray
        Base class for all genome arrays, array-like data structures mapping
        read counts to specific nucleotide positions.  
        
        Defines interfaces for:
            - retrieving vectors (as numpy.ndarray s) of read counts
              at each position of a GenomicSegment
              
            - getting the sum of read counts in a dataset
            
            - toggling normalization of fetched counts to reads-per-million
    
    MutableAbstractGenomeArray
        Base class for GenomeArray and SparseGenomeArray. Contains all interfaces
        from AbstractGenomeArray, and additionally defines interfaces for:
        
            - setting values, manually or mathematically, over regions of the
              genome, or over the entire genome, element wise
            
See also
--------
some_method
    A method that is useful

plastid.genomics.genome_array.GenomeArray
    A useful class
"""



import os
import numpy

from nose.tools import assert_less_equal, assert_raises, assert_dict_equal,\
                       assert_true, assert_equal, assert_almost_equal

from pkg_resources import resource_filename
from plastid.readers.bigwig import BigWigReader
from plastid.readers.wiggle import WiggleReader
from plastid.genomics.roitools import GenomicSegment
from plastid.genomics.genome_array import GenomeArray
from plastid.util.services.decorators import skip_if_abstract

TOL = 1e-5  #tolerance is high because bigwig files are approximate

base_path = resource_filename("plastid", "test/data")
wigfile = os.path.join(base_path, "command_line", "gen_reads_center_12_fw.wig")
bigwigfile = os.path.join(base_path, "command_line", "gen_reads_center_12_fw.bw")


class WildCard(float):
    def __eq__(self, other):
        return True

    def __neq__(self, other):
        return False

    def __repr__(self):
        return 'any float'

    def __str__(self):
        return 'any float'


wildcard = WildCard()


class AbstractTestBBIFile():
    @classmethod
    def setUpClass(cls):
        cls.bw = BigWigReader(bigwigfile, fill=0)

    @skip_if_abstract
    def test_chrom_sizes(self):
        found = self.bw.chroms
        expected = {}
        with open(os.path.join(base_path, "annotations", "sacCer3.sizes")) as fh:
            for line in fh:
                k, v = line.strip().split("\t")
                expected[k] = int(v)

        # these two happen not to be in the dataset
        expected.pop("chrVI")
        expected.pop("chrmt")

        assert_dict_equal(expected, found)

    @skip_if_abstract
    def test_no_crash_if_file_not_exist(self):
        with assert_raises(IOError) as _:
            _ = self.reader_class("non_existant_file")

    @skip_if_abstract
    def test_sum(self):
        assert False



class TestBigWigReader(AbstractTestBBIFile):
    @classmethod
    def setUpClass(cls):
        cls.bw = BigWigReader(bigwigfile, fill=0)
        cls.reader_class = BigWigReader
        cls.chrdict = {
            'chrI': 230218,
            'chrII': 813184,
            'chrIII': 316620,
            'chrIV': 1531933,
            'chrV': 576874,
            'chrVII': 1090940,
            'chrVIII': 562643,
            'chrIX': 439888,
            'chrX': 745751,
            'chrXI': 666816,
            'chrXII': 1078177,
            'chrXIII': 924431,
            'chrXIV': 784333,
            'chrXV': 1091291,
            'chrXVI': 948066
        }

    def check_vals_against_wig(self, expected, found):
        diff = abs(expected - found)
        maxdiff = diff.max()
        maxloc = diff.argmax()
        msg = "Maximum difference found between BigWig and Wiggle (%s) is at position %s and exceeded tolerance (%s).\n" % (
            maxdiff, maxloc, TOL
        )
        msg += "At that position, expected %s, got %s." % (expected[maxloc], found[maxloc])
        assert_less_equal(maxdiff, TOL, msg)

    # NOTE: this test relies on WiggleReader being correct
    def test_vals_against_wig(self):
        ga = GenomeArray()
        with open(wigfile) as fin:
            ga.add_from_wiggle(fin, "+")
            for chrom, length in self.chrdict.items():
                seg = GenomicSegment(chrom, 0, length, "+")
                expected = ga[seg]
                found = self.bw[seg]
                yield self.check_vals_against_wig, expected, found

    # NOTE: this test relies on WiggleReader being correct
    def check_random_windows_against_wig(self, strand):
        chrdict = self.chrdict
        chroms = list(self.chrdict)
        chridx = numpy.random.randint(0, high=len(chroms), size=50)
        ga = GenomeArray()

        i = 0

        with open(wigfile) as fin:
            ga.add_from_wiggle(fin, strand)
            while i < 50:
                chrom = chroms[chridx[i]]
                maxlength = chrdict[chrom]
                start = numpy.random.randint(0, high=maxlength - 2000)
                end = numpy.random.randint(start + 10000, high=start + 20000)

                # make sure we don't go off chrom
                while end > maxlength:
                    end = numpy.random.randint(start + 100, high=start + 10000)

                seg = GenomicSegment(chrom, start, end, strand)
                expected = ga[seg]
                # make sure segment has counts in it
                if expected.sum() > 0:
                    i += 1
                    found = self.bw[seg]
                    yield self.check_vals_against_wig, expected, found

    def test_random_windows_against_wig_fw(self):
        self.check_random_windows_against_wig("+")

    def test_random_windows_against_wig_rc(self):
        self.check_random_windows_against_wig("-")

    def test_get_chromosome_counts_zero_fill(self):
        ga = GenomeArray()
        with open(wigfile) as fin:
            ga.add_from_wiggle(fin, "+")
            for chrom, length in self.chrdict.items():
                seg = GenomicSegment(chrom, 0, length, "+")
                expected = ga[seg]
                found = self.bw.get_chromosome_counts(chrom)
                yield self.check_vals_against_wig, expected, found

    def test_fill_val_absent_chrom(self):
        filldef = BigWigReader(bigwigfile)
        fillnan = BigWigReader(bigwigfile, fill=numpy.nan)
        fill0 = BigWigReader(bigwigfile, fill=0)
        fill10 = BigWigReader(bigwigfile, fill=10)

        # chrVI is not in dataset; this should be an empty array
        seg = GenomicSegment("chrVI", 5, 1000, "+")

        assert_equal(len(filldef[seg]), len(seg), "fetched wrong size")

        #         assert_true(numpy.isnan(filldef[seg]).all(),
        #                     "default not nan")
        #
        #         assert_true(numpy.isnan(fillnan[seg]).all(),
        #                     "nanfill didn't work")

        assert_true((fill0[seg] == 0).all(), "0-fill didn't work")

#         assert_true((fill10[seg] == 10).all(),
#                     "10-fill didn't work")

    def test_fill_val_present_chrom(self):
        filldef = BigWigReader(bigwigfile)
        fillnan = BigWigReader(bigwigfile, fill=numpy.nan)
        fill0 = BigWigReader(bigwigfile, fill=0)
        fill10 = BigWigReader(bigwigfile, fill=10)

        # empty region
        seg = GenomicSegment("chrIV", 5, 10, "+")

        assert_equal(len(filldef[seg]), len(seg), "fetched wrong size")

        #         assert_true(numpy.isnan(filldef[seg]).all(),
        #                     "default not nan")
        #
        #         assert_true(numpy.isnan(fillnan[seg]).all(),
        #                     "nanfill didn't work")

        assert_true((fill0[seg] == 0).all(), "0-fill didn't work")

#         assert_true((fill10[seg] == 10).all(),
#                     "10-fill didn't work")

    def test_sum(self):
        bw = BigWigReader(os.path.join(base_path, "mini", "wig", "bw_fiveprime_15_fw.bw"))
        assert_equal(bw.sum(), 4000)

    def test_iter(self):
        with open(wigfile) as fh:
            wig = WiggleReader(fh)
            bw = BigWigReader(bigwigfile)

            for found in bw:
                expected = next(wig)
                fchrom = found[0]
                echrom = expected[0]
                assert_equal(
                    fchrom, echrom,
                    "Chromosome mismatch. Expected '%s', found '%s'." % (fchrom, echrom)
                )

                fstart = found[1]
                estart = expected[1]
                assert_equal(
                    fstart, estart, "Start mismatch. Expected '%s', found '%s'." % (fstart, estart)
                )

                fend = found[2]
                eend = expected[2]
                assert_equal(fend, eend, "End mismatch. Expected '%s', found '%s'." % (fend, eend))

                fval = found[3]
                eval_ = expected[3]
                diff = abs(fval - eval_)
                assert_true(
                    diff < TOL, "Difference %s exceeds tolerance '%s'. Expected '%s', found '%s'." %
                    (diff, TOL, fval, eval_)
                )


# Disabled until we decide what to do with summarize()
#     def test_summarize(self):
#         bw  = BigWigReader(bigwigfile)
#
#         #ga = GenomeArray(bw.chroms)
#         #ga.add_from_wiggle(open(wigfile),"+")
#
#         chrom = "chrI"
#         maxlen = bw.chroms[chrom]
#         winstarts = numpy.random.randint(0,maxlen-20000,size=10)
#         winends   = winstarts + numpy.random.randint(500,40000,size=10)
#         winends[winends > maxlen] = maxlen
#
#         numtests = 10
#         i = 0
#         while i < numtests:
#             s = numpy.random.randint(0,high=maxlen)
#             e = min(maxlen,numpy.random.randint(s+500,s+40000))
#             seg = GenomicSegment(chrom,s,e,"+")
#             #arr = ga[seg]
#             arr = bw[seg]
#
#             labels = ["mean","max","min","cov","stdev"]
#             expected = [arr.mean(),arr.max(),arr.min(),wildcard,arr.std()]
#
#             # change nans to 0
#             expected = [0 if numpy.isnan(X) else X for X in expected]
#
#             print(expected)
#             found = bw.summarize(seg)
#             print(found)
#             print("---------------")
#             for label, exval, fval in zip(labels,expected,found):
#                 msg =  "test_summarize failed for stat '%s'. Expected %s, got %s (diff: %s)." % (label,exval,fval,abs(exval-fval))
#                 assert_almost_equal(exval,fval,msg=msg,delta=5)
#
#             i += 1
#
#         # retval for summarize:  (mean,max_,min_,cov,stdev)



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.psite`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.util.services.decorators import catch_stderr
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import REF_FILES
from plastid.bin.psite import main

#===============================================================================
# INDEX: global constants
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.psite",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="psite"),
}

_outbase = os.path.join(test_info["temp_file_path"], "test_psite")

#===============================================================================
# INDEX: tests
#===============================================================================

psite_tests = [
    (
        "%s %s --count_files %s --norm_region 70 150 --require_upstream --min_length 26 --max_length 31"
        % (
            REF_FILES["yeast_metagene_cds_start"],
            _outbase,
            REF_FILES["yeast_rp_bam"],
        ), [REF_FILES["yeast_psite"]], [_outbase + "_p_offsets.txt"], [""]
    ),
    (
        "%s %s --count_files %s --norm_region 70 150 --require_upstream --min_length 26 --max_length 31 --constrain 0 11"
        % (
            REF_FILES["yeast_metagene_cds_start"],
            _outbase,
            REF_FILES["yeast_rp_bam"],
        ), [REF_FILES["yeast_psite_constrain"]], [_outbase + "_p_offsets.txt"], [""]
    ),
    # test using --normalize_over instead of --norm_region
    (
        "%s %s --count_files %s --normalize_over 20 100 --require_upstream --min_length 26 --max_length 31"
        % (
            REF_FILES["yeast_metagene_cds_start"],
            _outbase,
            REF_FILES["yeast_rp_bam"],
        ), [REF_FILES["yeast_psite"]], [_outbase + "_p_offsets.txt"], [""]
    ),
    # test using --normalize_over instead of --norm_region
    (
        "%s %s --count_files %s --normalize_over 20 100 --require_upstream --min_length 26 --max_length 31 --constrain 0 11"
        % (
            REF_FILES["yeast_metagene_cds_start"],
            _outbase,
            REF_FILES["yeast_rp_bam"],
        ), [REF_FILES["yeast_psite_constrain"]], [_outbase + "_p_offsets.txt"], [""]
    ),
]

# /home/joshua/projects/plastid/plastid/test/data/command_line/gen_cds_start_rois.txt /tmp/psite5SY5Wc/test_psite --count_files /home/joshua/projects/plastid/plastid/test/data/command_line/gen_reads.bam --norm_region 70 150 --require_upstream --min_length 26 --max_length 31
"""Functional tests of :py:mod:`plastid.bin.psite`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""

#===============================================================================
# INDEX: test/helper functions
#===============================================================================


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for :py:mod:`plastid.bin.psite`"""
    for x in execute_helper(test_info, psite_tests):
        yield x



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.test_table_equality`"""
import os
import shlex
import numpy
import pandas as pd
from random import shuffle
from tempfile import NamedTemporaryFile
from pkg_resources import cleanup_resources
from nose.plugins.attrib import attr
from nose.tools import assert_equal
from plastid.util.io.filters import CommentWriter
from plastid.util.io.openers import NullWriter
from plastid.genomics.roitools import GenomicSegment
from plastid.bin.test_table_equality import main
from plastid.util.services.decorators import catch_stderr

size = 5000

#capture output
run_main = catch_stderr()(main)


@attr(test="functional")
def test_exit_status():

# yapf: disable
    # define columns
    cols = {
        "intA"   : numpy.random.randint(0, high=2**16, size=size),
        "intB"   : numpy.random.randint(-10, high=20, size=size),
        "idxA"   : numpy.arange(size),
        "chrA"   : numpy.array([chr(65+(X%(91-65))) for X in range(size)]),
        "strA"   : numpy.array([str(GenomicSegment("chrA", X, X+500, "+")) for X in range(size)]),
        "strB"   : numpy.array([str(GenomicSegment("chrB", X/2, X/2+500, "-")) for X in range(size)]),
        "floatA" : 10*numpy.random.randn(size) + 500,
        "floatB" : (10**-5)*numpy.random.random(size),
        "objA"   : numpy.tile(None, 5000),
        "objB"   : numpy.array([GenomicSegment("chrC", X, X+Y, "+") for X, Y in zip(range(size), numpy.random.randint(2, high=1000, size=size))]),
          }

    # allocate temp files we will use
    headerfile                        = NamedTemporaryFile(delete=False, mode="w")
    headerfile_extra_cols             = NamedTemporaryFile(delete=False, mode="w")
    headerfile_extra_cols_diff        = NamedTemporaryFile(delete=False, mode="w")
    headerfile_extra_cols_shuffled    = NamedTemporaryFile(delete=False, mode="w")
    headerfile_shuffled               = NamedTemporaryFile(delete=False, mode="w")
    headerfile_diff_vals              = NamedTemporaryFile(delete=False, mode="w")

    noheaderfile                      = NamedTemporaryFile(delete=False, mode="w")
    noheaderfile_extra_cols           = NamedTemporaryFile(delete=False, mode="w")
    noheaderfile_extra_cols_diff      = NamedTemporaryFile(delete=False, mode="w")
    noheaderfile_extra_cols_shuffled  = NamedTemporaryFile(delete=False, mode="w")
    noheaderfile_shuffled             = NamedTemporaryFile(delete=False, mode="w")
    noheaderfile_diff_vals            = NamedTemporaryFile(delete=False, mode="w")

    # write values
    keyorder = ["idxA"] + sorted(list(set(cols.keys()) - {"idxA"}))

    table1 = pd.DataFrame(cols)
    table1.to_csv(headerfile,   index=False, header=True, sep="\t")
    table1.to_csv(noheaderfile, index=False, header=False, sep="\t", columns=keyorder)
    headerfile.close()
    noheaderfile.close()
# yapf: enable

    table1["extra"] = 2**7 * numpy.random.random(size=size)
    table1.to_csv(headerfile_extra_cols, index=False, header=True, sep="\t")
    table1.to_csv(
        noheaderfile_extra_cols, index=False, header=False, sep="\t", columns=["extra"] + keyorder
    )

    headerfile_extra_cols.close()
    noheaderfile_extra_cols.close()

    table1["extra"] += 10**-4 * numpy.random.random(size=size)
    table1.to_csv(headerfile_extra_cols_diff, index=False, header=True, sep="\t")
    table1.to_csv(
        noheaderfile_extra_cols_diff,
        index=False,
        header=False,
        sep="\t",
        columns=["extra"] + keyorder
    )

    headerfile_extra_cols_diff.close()
    noheaderfile_extra_cols_diff.close()

    shufidx = numpy.arange(size)
    shuffle(shufidx)
    table2 = pd.DataFrame({K: V[shufidx] for K, V in cols.items()})
    table2.to_csv(headerfile_shuffled, index=False, header=True, sep="\t")
    table2.to_csv(noheaderfile_shuffled, index=False, header=False, sep="\t", columns=keyorder)
    headerfile_shuffled.close()
    noheaderfile_shuffled.close()

    table2["extra"] = table1["extra"][shufidx]
    table2.to_csv(headerfile_extra_cols_shuffled, index=False, header=True, sep="\t")
    table2.to_csv(
        noheaderfile_extra_cols_shuffled,
        index=False,
        header=False,
        sep="\t",
        columns=["extra"] + keyorder
    )

    headerfile_extra_cols_shuffled.close()
    noheaderfile_extra_cols_shuffled.close()

    # Define tests, as tuples of:
    #   -Test name/description
    #   -Command-line arguments to pass to :py:mod:`plastid.bin.test_table_equality`
    #   -Expected exit code/returns status for :py:func:`main`
    tests = [
        ("same", "%s %s" % (headerfile.name, headerfile.name), 0),
        ("diff_column_names", "%s %s" % (headerfile.name, headerfile_extra_cols.name), 1),
        (
            "extra_column_names_ignored",
            "%s %s --exclude extra" % (headerfile.name, headerfile_extra_cols.name), 0
        ), ("shuffled_rows", "%s %s" % (headerfile.name, headerfile_shuffled.name), 1),
        (
            "shuffled_rows_name_sort",
            "%s %s --sort_keys idxA" % (headerfile.name, headerfile_shuffled.name), 0
        ),
        (
            "shuffled_rows_multi_name_sort",
            "%s %s --sort_keys strB chrA" % (headerfile.name, headerfile_shuffled.name), 0
        ),
        (
            "same_column_names_diff_values",
            "%s %s" % (headerfile_extra_cols.name, headerfile_extra_cols_diff.name), 1
        ),
        (
            "same_column_names_diff_values_tol",
            "%s %s --tol 0.01" % (headerfile_extra_cols.name, headerfile_extra_cols_diff.name), 0
        ),
        (
            "same_column_names_diff_values_ignored",
            "%s %s --exclude extra" % (headerfile_extra_cols.name, headerfile_extra_cols_diff.name), 0
        ),
        (
            "shuffled_rows_extra_columns_ignored",
            "%s %s --exclude extra" % (headerfile.name, headerfile_extra_cols_shuffled.name), 1
        ),
        (
            "shuffled_rows_extra_columns_ignored_name_sort",
            "%s %s --exclude extra --sort_keys idxA" %
            (headerfile.name, headerfile_extra_cols_shuffled.name), 0
        ),
        ("noheader_same", "%s %s --no_header" % (noheaderfile.name, noheaderfile.name), 0),
        (
            "noheader_extra_columns",
            "%s %s --no_header" % (noheaderfile.name, noheaderfile_extra_cols.name), 1
        ),
        (
            "noheader_shuffled_rows",
            "%s %s --no_header" % (noheaderfile.name, noheaderfile_shuffled.name), 1
        ),
        (
            "noheader_shuffled_rows_int_sort",
            "%s %s --no_header --sort_keys 0" % (noheaderfile.name, noheaderfile_shuffled.name), 0
        ),
        (
            "noheader_diff_values",
            "%s %s --no_header" % (noheaderfile_extra_cols.name, noheaderfile_extra_cols_diff.name),
            1
        ),
        (
            "noheader_diff_values_tol", "%s %s --no_header --tol 0.01" %
            (noheaderfile_extra_cols.name, noheaderfile_extra_cols_diff.name), 0
        ),
        (
            "no_header_diff_values_ignored", "%s %s --no_header --exclude 0" %
            (noheaderfile_extra_cols.name, noheaderfile_extra_cols_diff.name), 0
        ),
        (
            "no_header_shuffled_rows_extra_columns_ignored_int_sort",
            "%s %s --no_header --exclude 0 --sort_keys 1" %
            (noheaderfile_extra_cols.name, noheaderfile_extra_cols_shuffled.name), 0
        )
    ] # yapf: disable
    """ Tests to conduct, as tuples of:

        - Test name/description
        - Command-line arguments to pass to :py:mod:`plastid.bin.test_table_equality`
        - Expected exit code/returns status for :py:func:`main`
    """
    for test_name, argstr, expected_exit in tests:
        yield check_exit_status, test_name, argstr, expected_exit

    # clean up
    os.unlink(headerfile.name)
    os.unlink(headerfile_extra_cols.name)
    os.unlink(headerfile_extra_cols_diff.name)
    os.unlink(headerfile_extra_cols_shuffled.name)
    os.unlink(headerfile_shuffled.name)
    os.unlink(headerfile_diff_vals.name)

    os.unlink(noheaderfile.name)
    os.unlink(noheaderfile_extra_cols.name)
    os.unlink(noheaderfile_extra_cols_diff.name)
    os.unlink(noheaderfile_extra_cols_shuffled.name)
    os.unlink(noheaderfile_shuffled.name)
    os.unlink(noheaderfile_diff_vals.name)
    cleanup_resources()


def check_exit_status(test_name, argstr, expected_exit):
    """Verify exit status from :py:mod:`~plastid.bin.test_table_equality`

    Parameters
    ----------
    test_name : str
        Name of test (for reporting purposes)

    argstr : str
        Command-line arguments to pass to :py:mod:`~plastid.bin.test_table_equality`

    expected_exit : int
        Expected exit status
    """
    found_exit, failures = run_main(shlex.split(argstr), verbose=True)
    msg = "    \n".join(failures)
    assert_equal(found_exit, expected_exit,
                 "Failed test %s. Got exit code %s,  expected %s.\n    %s" % \
                         (test_name, found_exit, expected_exit, msg))



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.util.scriptlib.argparsers`"""
import unittest
import shlex
import argparse
import copy
import numpy
import itertools
import warnings
from pkg_resources import cleanup_resources
from twobitreader import TwoBitFile
from nose.plugins.attrib import attr
from nose.tools import (
    assert_equal,
    assert_greater,
    assert_raises,
    assert_in,
    assert_not_in,
    assert_true,
    assert_list_equal,
)

from plastid.test.ref_files import MINI, REF_FILES, EXTENSIONS
from plastid.util.io.filters import CommentReader
from plastid.readers.bed import BED_Reader
from plastid.util.services.exceptions import MalformedFileError
from plastid.genomics.roitools import SegmentChain, Transcript
from plastid.genomics.map_factories import (
    FivePrimeMapFactory,
    ThreePrimeMapFactory,
    CenterMapFactory,
    VariableFivePrimeMapFactory,
)

from plastid.genomics.genome_array import (
    GenomeArray,
    BAMGenomeArray,
    SparseGenomeArray,
    BigWigGenomeArray,
)

from plastid.genomics.genome_hash import GenomeHash, BigBedGenomeHash

from plastid.util.scriptlib.argparsers import (
    PrefixNamespaceWrapper,
    get_alignment_file_parser,
    get_genome_array_from_args,
    get_annotation_file_parser,
    get_transcripts_from_args,
    get_segmentchain_file_parser,
    get_segmentchains_from_args,
    get_mask_file_parser,
    get_genome_hash_from_mask_args,
    get_sequence_file_parser,
    get_seqdict_from_args,
    _parse_variable_offset_file,
)

from plastid.util.services.mini2to3 import StringIO

warnings.simplefilter("ignore")


def tearDownModule():
    cleanup_resources()


@attr(test="unit")
class TestParseVariableOffsetFile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ref_file = REF_FILES["yeast_psite"]
        with open(cls.ref_file) as fh:
            cls.lines = CommentReader(fh).read().strip("\n").split("\n")
        cls.expected = {
            26: 12,
            27: 12,
            28: 13,
            29: 13,
            30: 14,
            31: 13,
            "default": 13,
        }

    def test_no_default(self):
        my_expected = copy.deepcopy(self.expected)
        my_expected.pop("default")
        my_lines = "\n".join(self.lines[:-1])
        dtmp = _parse_variable_offset_file(StringIO.StringIO(my_lines))
        self.assertDictEqual(my_expected, dtmp)

    def test_with_default(self):
        with open(self.ref_file) as fh:
            dtmp = _parse_variable_offset_file(CommentReader(fh))
        self.assertDictEqual(self.expected, dtmp)

    def test_only_default(self):
        dtmp = _parse_variable_offset_file(StringIO.StringIO(self.lines[-1]))
        my_expected = {"default": 13}
        self.assertDictEqual(my_expected, dtmp)

    def test_redefined_raises_exception(self):
        # make sure files with multiple entries for the same key error out
        my_lines = "\n".join(self.lines + self.lines)
        fakefile = StringIO.StringIO(my_lines)
        fakefile.__name__ = "Fake file"
        self.assertRaises(MalformedFileError, _parse_variable_offset_file, fakefile)

    def test_wrong_cols_raises_exception(self):
        too_many = """25    12
26    13
27    14
28    15    52
""".replace("    ", "\t")
        fakefile = StringIO.StringIO(too_many)
        fakefile.__name__ = "Fake file"
        self.assertRaises(MalformedFileError, _parse_variable_offset_file, fakefile)

        too_few = """25    12
26    13
27
28    15
""".replace("    ", "\t")
        fakefile = StringIO.StringIO(too_few)
        fakefile.__name__ = "Fake file"
        self.assertRaises(MalformedFileError, _parse_variable_offset_file, fakefile)

    def test_non_int_raises_exception(self):
        non_int_key = """25    12
26    13
a    14
28    15
""".replace("    ", "\t")
        fakefile = StringIO.StringIO(non_int_key)
        fakefile.__name__ = "Fake file"
        self.assertRaises(MalformedFileError, _parse_variable_offset_file, fakefile)

        non_int_val = """25    12
26    13
27    q
28    15
""".replace("    ", "\t")
        fakefile = StringIO.StringIO(non_int_val)
        fakefile.__name__ = "Fake file"
        self.assertRaises(MalformedFileError, _parse_variable_offset_file, fakefile)

        null_val = """25    12
26    13
27    
28    15
""".replace("    ", "\t")
        fakefile = StringIO.StringIO(null_val)
        fakefile.__name__ = "Fake file"
        self.assertRaises(MalformedFileError, _parse_variable_offset_file, fakefile)


#=============================================================================
# INDEX: constants and helper functions used for tests in multiple sections
#        below.
#=============================================================================

offset_file = REF_FILES["yeast_psite"]
bamfilename = MINI["bamfile"]
bowtiefilename = MINI["bowtie_file"]
wigglefilename = MINI["wig_bedgraph"]
bigwigfilename = MINI["bigwig"]
with open(MINI["bed_file"]) as fh:
    regions = list(BED_Reader(fh, return_type=SegmentChain))

spliced_regions = [X for X in regions if len(X) > 1]
unique_regions  = [X for X in regions if "repeat" not in X.get_name()] 
introns = [X for X in regions if "intron" not in X.get_name()]

#unique_regions - spliced_regions - introns
unspliced_unique_regions = [
    X for X in unique_regions if all([len(X) == 1, "intron" not in X.get_name()])
]  

assert_greater(len(unspliced_unique_regions), 0)

alignment_file_parser_disableable = (
    "fiveprime", "threeprime", "center", "offset", "nibble", "min_length", "max_length",
    "normalize", "big_genome",
)
alignment_file_parser_opts = (
    "big_genome", "count_files", "countfile_format", "mapping", "max_length", "min_length",
    "nibble", "normalize", "offset",
)

annotation_file_parser_opts = {"annotation_format", "add_three", "annotation_files", "tabix"}
annotation_file_gff_opts = {"gff_exon_types", "gff_transcript_types"}
annotation_file_parser_disableable = annotation_file_parser_opts - set(["annotation_files"])

ivcollection_file_parser_opts = annotation_file_parser_opts - set(["add_three"])
ivcollection_file_parser_disableable = annotation_file_parser_disableable - set(["add_three"])

mask_file_parser_opts = ivcollection_file_parser_opts
mask_file_parser_disableable = ivcollection_file_parser_disableable

alignment_file_parser = get_alignment_file_parser()
annotation_file_parser = get_annotation_file_parser()
ivcollection_file_parser = get_segmentchain_file_parser()
sequence_file_parser = get_sequence_file_parser()
mask_file_parser = get_mask_file_parser()


def check_prefix(parser_fn, opts):
    """Helper function to test prefix appending to various parsers

    Parameters
    ----------
    parser_fn : function
        Function that returns a :py:class:`~argparse.ArgumentParser`

    opts : list
        list of strings of options, without prefixes, included
        in the parser
    """
    parser = parser_fn(prefix="test")
    args = parser.parse_args([])
    for opt in opts:
        # getattr raises an AttributeError if attribute not present
        # so it is a pseudotest
        getattr(args, "%s%s" % ("test", opt))


def check_disabled_options(parser_fn, disable_opts):
    """Helper function to test disabling of options in various parsers

    Parameters
    ----------
    parser_fn : function
        Function that returns a :py:class:`~argparse.ArgumentParser`

    disable_opts : list
        list of strings of options, without prefixes, that can be 
        disabled in the parser
    """
    # kind of a hack. make sure options are disabled
    # by asserting that they don't appear in --help
    for d in disable_opts:
        parser = parser_fn(disabled=[d])
        assert_not_in("\n  --%s " % d, parser.format_help())


@attr(test="unit")
def test_parser_prefixes():
    tuples = [
        (get_alignment_file_parser, alignment_file_parser_opts),
        (get_annotation_file_parser, annotation_file_parser_opts),
        (get_segmentchain_file_parser, ivcollection_file_parser_opts),
        (get_mask_file_parser, mask_file_parser_opts),
    ]
    for fn, opts in tuples:
        yield check_prefix, fn, opts


@attr(test="unit")
def test_parser_disabled_options():
    tuples = [
        (get_alignment_file_parser, alignment_file_parser_disableable),
        (get_annotation_file_parser, annotation_file_parser_disableable),
        (get_segmentchain_file_parser, ivcollection_file_parser_disableable),
        (get_mask_file_parser, mask_file_parser_disableable),
    ]
    for fn, opts in tuples:
        yield check_disabled_options, fn, opts


#=============================================================================
# INDEX: tests for alignment file parsing
#=============================================================================


@attr(test="unit")
def test_alignment_parser_no_count_file_causes_exit():
    args = alignment_file_parser.parse_args([])
    assert_raises(SystemExit, get_genome_array_from_args, args)


@attr(test="unit")
def test_alignment_parser_no_mapping_causes_exit():
    argstr = "--countfile_format BAM --count_files %s" % bamfilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    assert_raises(SystemExit, get_genome_array_from_args, args)

    argstr = "--countfile_format bowtie --count_files %s" % bamfilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    assert_raises(SystemExit, get_genome_array_from_args, args)


@attr(test="unit")
def test_alignment_parser_open_bam_makes_bamgenomearray():
    argstr = "--countfile_format BAM --count_files %s --fiveprime" % bamfilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))

    assert_equal(bamfilename, args.count_files[0])

    ga = get_genome_array_from_args(args)
    assert_true(isinstance(ga, BAMGenomeArray))


@attr(test="unit")
def test_alignment_parser_open_bigwig_makes_bigwiggenomearray():
    argstr = "--countfile_format bigwig --count_files %s" % bigwigfilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    assert_true(isinstance(ga, BigWigGenomeArray))


@attr(test="unit")
def test_alignment_parser_open_bowtie_makes_genomearray():
    argstr = "--countfile_format bowtie --count_files %s --fiveprime" % bowtiefilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    assert_true(isinstance(ga, GenomeArray))


@attr(test="unit")
def test_alignment_parser_open_wiggle_makes_genomearray():
    argstr = "--countfile_format wiggle --count_files %s" % wigglefilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    assert_true(isinstance(ga, GenomeArray))


@attr(test="unit")
def test_alignment_parser_bowtie_with_sparsegenomearray():
    argstr = "--countfile_format bowtie --count_files %s --fiveprime --big_genome" % bowtiefilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    assert_true(isinstance(ga, SparseGenomeArray))


@attr(test="unit")
def test_alignment_parser_wiggle_with_sparsegenomearray():
    argstr = "--countfile_format wiggle --count_files %s --big_genome" % wigglefilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))

    ga = get_genome_array_from_args(args)
    assert_true(isinstance(ga, SparseGenomeArray))


def check_genome_array_against_vector(ga, mapping, offset_or_nibble):
    """Compare values in a |GenomeArray| against known vectors
    across a set of unspliced, uniquely mapping regions

    Parameters
    ----------
    ga : |GenomeArray|, |SparseGenomeArray|, or |BAMGenomeArray|
        Genome array to test

    mapping : str
        String specifying mapping rule. "fiveprime", "threeprime",
        "center", or "fiveprime_variable"

    offset_or_nibble : int or str
        Integer specifying offset or nibble, or str specifying filename
        of variable offset dict
    """
    if mapping != "fiveprime_variable":
        vec_fw = numpy.loadtxt(MINI["vec_%s_%s_fw.txt" % (mapping, offset_or_nibble)])
        vec_rc = numpy.loadtxt(MINI["vec_%s_%s_rc.txt" % (mapping, offset_or_nibble)])
    else:
        vec_fw = numpy.loadtxt(MINI["vec_threeprime_15_fw.txt"])
        vec_rc = numpy.loadtxt(MINI["vec_threeprime_15_rc.txt"])

    for region in unspliced_unique_regions:
        vec = []
        if region.strand == "+":
            for iv in region:
                vec.extend(vec_fw[iv.start:iv.end])
        else:
            for iv in region:
                vec.extend(vec_rc[iv.start:iv.end])
            vec = vec[::-1]

        vec = numpy.array(vec)
        vec2 = numpy.array(region.get_counts(ga))
        diff = abs(vec2 - vec)
        assert_true(
            (diff <= 1e-8).all(), "Maximum difference exceeded on %s: %s\n%s\n%s" %
            (region.get_name(), diff.max(), vec[:40], vec2[:40])
        )


def check_bam_mapping(mapping, offset_or_nibble):
    """Generator function to test import of BAM files under many read mapping rules

    Parameters
    ----------
    mapping : str
        String specifying mapping rule. "fiveprime",
        "threeprime", "center", or "fiveprime_variable"

    offset_or_nibble : int or str
        Integer specifying offset or nibble (only for center mapping)
        or str specifying filename of variable offset dict text file
    """
    argstr = "--countfile_format BAM --count_files %s --%s " % (bamfilename, mapping)
    if mapping == "center":
        argstr += ("--nibble %s" % offset_or_nibble)
    else:
        argstr += ("--offset %s" % offset_or_nibble)

    mapdict = {
        "fiveprime": FivePrimeMapFactory,
        "threeprime": ThreePrimeMapFactory,
        "center": CenterMapFactory,
        "fiveprime_variable": VariableFivePrimeMapFactory,
    }

    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    #assert_equal(mapping,ga.map_fn.__mapping__)a

    assert_true(isinstance(ga.map_fn, mapdict[mapping]))

    # assure correct mapping function was applied
    # by comparing counts loaded from a bowtie file
    # under those map rules to a vector corresponding
    # to known answers for those rules
    assert_greater(ga.sum(), 0)
    check_genome_array_against_vector(ga, mapping, offset_or_nibble)


def check_bowtie_mapping(mapping, offset_or_nibble):
    """Generator function to test import of bowtie files under many read mapping rules

    Parameters
    ----------
    mapping : str
        String specifying mapping rule. "fiveprime",
        "threeprime", "center", or "fiveprime_variable"

    offset_or_nibble : int or str
        Integer specifying offset or nibble (only for center mapping)
        or str specifying filename of variable offset dict text file
    """
    # assure correct mapping function was applied
    # by comparing counts loaded from a bowtie file
    # under those map rules to a vector corresponding
    # to known answers for those rules
    argstr = "--countfile_format bowtie --count_files %s --%s " % (bowtiefilename, mapping)
    if mapping == "center":
        argstr += ("--nibble %s" % offset_or_nibble)
    else:
        argstr += ("--offset %s" % offset_or_nibble)

    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    assert_greater(ga.sum(), 0)
    check_genome_array_against_vector(ga, mapping, offset_or_nibble)


@attr(test="unit")
def test_alignment_parser_bam_mapping():
    for mapping in ("fiveprime", "threeprime", "center"):
        yield check_bam_mapping, mapping, 0

    yield check_bam_mapping, "fiveprime_variable", offset_file


@attr(test="unit")
def test_alignment_parser_bowtie_mapping():
    for mapping in ("threeprime", "fiveprime"):
        for offset in (0, 15):
            yield check_bowtie_mapping, mapping, offset

    for nibble in (0, 12):
        yield check_bowtie_mapping, "center", nibble

    yield check_bowtie_mapping, "fiveprime_variable", offset_file


@attr(test="unit")
def test_alignment_parser_bam_fiveprime_variable_requires_offset():
    argstr = "--countfile_format BAM --count_files %s --fiveprime_variable" % bamfilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    assert_raises(SystemExit, get_genome_array_from_args, args)


@attr(test="unit")
def test_alignment_parser_bowtie_fiveprime_varible_requires_offset():
    argstr = "--countfile_format bowtie --count_files %s --fiveprime_variable" % bowtiefilename
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    assert_raises(SystemExit, get_genome_array_from_args, args)


def alignment_parser_check_normalize(fmt, filename):
    argstr = "--countfile_format %s --count_files %s --fiveprime --normalize" % (fmt, filename)
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    assert_true(ga._normalize)


@attr(test="unit")
def test_alignment_parser_normalize():
    tups = [
        ("wiggle", wigglefilename),
        ("BAM", bamfilename),
        ("bowtie", bowtiefilename),
        ("bigwig", bigwigfilename)
    ]
    for fmt, filename in tups:
        yield alignment_parser_check_normalize, fmt, filename


@attr(test="unit")
def test_alignment_parser_input_choices():
    # kind of a hack. make sure input choices
    # are correct based upon whether or not they
    # appear in --help
    all_choices = ("bowtie", "tagalign", "wiggle", "BAM", "bigwig")
    for chosen in itertools.chain.from_iterable(
            itertools.combinations(all_choices, K) for K in range(len(all_choices))):
        tmp_str = "  --countfile_format {%s}" % ",".join(list(chosen))
        parser = get_alignment_file_parser(input_choices=chosen)
        assert_in(tmp_str, parser.format_help())


@attr(test="unit")
def test_alignment_parser_set_sum():
    tups = [
        ("wiggle", wigglefilename),
        ("BAM", bamfilename),
        ("bowtie", bowtiefilename),
        ("bigwig", bigwigfilename)
    ]
    for fmt, filename in tups:
        yield alignment_parser_check_set_sum, fmt, filename


def alignment_parser_check_set_sum(fmt, filename):
    my_sum = 5
    argstr = "--countfile_format %s --count_files %s --fiveprime --sum %s" % (fmt, filename, my_sum)
    args = alignment_file_parser.parse_args(shlex.split(argstr))
    ga = get_genome_array_from_args(args)
    assert_equal(my_sum, ga.sum())


#=============================================================================
# INDEX: tests for annotation file parsing
#=============================================================================


def check_annotation_input_choices(parser_fn):
    """Check that correct input choices are offered for various annotation file parsers
    and that GFF3-specific options are offered only when GFF3 is an option

    Parameters
    ----------
    parser_fn : function
        Function that returns an :py:class:`~argparse.ArgumentParser`
        of genome annotation files
    """
    all_choices = ["BED", "BigBed", "GTF2", "GFF3", "PSL"]
    for chosen in itertools.chain.from_iterable(
            itertools.combinations(all_choices, K) for K in range(len(all_choices))):
        tmp_str = "  --annotation_format {%s}" % (",".join(list(chosen)))
        parser = parser_fn(input_choices=chosen)
        assert_in(tmp_str, parser.format_help())

        for attr in annotation_file_gff_opts:
            args = parser.parse_args([])
            if "GFF3" in chosen:
                getattr(args, attr)
            else:
                assert_raises(AttributeError, getattr, args, attr)


@attr(test="unit")
def test_annotation_input_choices():
    pfuncs = [
        get_annotation_file_parser,
        get_segmentchain_file_parser,
    ]
    for parser_fn in pfuncs:
        yield check_annotation_input_choices, parser_fn


def check_get_ivcollection_from_args_return_type(parser, func, return_type, fmt):
    """Check return type of annotation file reading

    Parameters
    ----------
    parser : :py:class:`argparse.ArgumentParser`
        Argument parser for annotation files

    func : function
        Function that parsers the :py:class:`~argparse.ArgumentParser`
        into lists of accepted and rejected features of interest

    return_type : class
        Return type expected by function. |SegmentChain|, |Transcript|,
        or |BLAT_Alignment|

    fmt : str
        Format of input file ("BED", "BigBed", "GTF2", "GFF3", or "PSL")
    """
    argstr = "--annotation_format %s --annotation_files %s" % (fmt, MINI["%s_file" % fmt.lower()])
    args = parser.parse_args(shlex.split(argstr))
    transcripts = func(args)
    for tx in transcripts:
        assert_true(isinstance(tx, return_type))


@attr(test="unit")
def test_get_ivcollection_from_args_return_type():
    for fmt in ("BED", "BigBed", "GTF2", "GFF3", "PSL"):
        yield check_get_ivcollection_from_args_return_type, ivcollection_file_parser, get_segmentchains_from_args, SegmentChain, fmt


@attr(test="unit")
def test_get_transcript_from_args_return_type():
    for fmt in ("BED", "BigBed", "GTF2", "GFF3"):
        yield check_get_ivcollection_from_args_return_type, annotation_file_parser, get_transcripts_from_args, Transcript, fmt


def check_get_transcript_from_args_add_three(parser, fmt, add_three, n=1):
    """Make sure stop codons are moved by three if ``add_three`` is specified
    in argument string and not moved by three if not specified.
    These are checked against expected values found in a BED file.
    
    Parameters
    ----------
    parser : :py:class:`argparse.ArgumentParser`
        Argument parser for annotation files

    fmt : str
        Format of input file ("BED", "BigBed", "GTF2", "GFF3", or "PSL")
        
    add_three : bool
        Whether or not to move stop codons by three nucleotides
    
    n : int
        Number of times input file was passed to parser (Default : 1)
    """
    # GTF file does not include stop codon; others do.
    # So, GTF file will be off by three relative to others
    # unless it includes explicit stop codon features
    #
    # we correct for that here
    gtf_offset = 3 if fmt == "GTF2" else 0
    bed_ref = MINI["bed_file"]
    dtmp = {}
    with open(bed_ref) as fh:
        for line in fh:
            items = line.strip("\n").split("\t")
            name = items[3]
            strand = items[5]
            cds_start = int(items[6])
            cds_end = int(items[7])
            if cds_start < cds_end:
                if strand == "+":
                    cds_end -= gtf_offset
                else:
                    cds_start += gtf_offset

                if add_three == True:
                    if strand == "+":
                        cds_end += 3
                    else:
                        cds_start -= 3
                dtmp[name] = (cds_start, cds_end)
            else:
                dtmp[name] = (None, None)

    files = " ".join([MINI["%s_file" % fmt.lower()]] * n)
    argstr = "--annotation_format %s --annotation_files %s " % (fmt, files)
    if add_three == True:
        argstr += " --add_three"

    args = parser.parse_args(shlex.split(argstr))
    transcripts = list(get_transcripts_from_args(args))
    for tx in transcripts:
        name = tx.get_name()
        expected_start = dtmp[name][0]
        expected_end = dtmp[name][1]
        found_start = tx.cds_genome_start
        found_end = tx.cds_genome_end
        if found_start is None:
            assert_true(found_end is None)
            assert_true(
                expected_start is None, "Unequal CDS start (expected %s, got %s) on transcript %s" %
                (expected_start, found_start, name)
            )
            assert_true(
                expected_end is None, "Unequal CDS stop (expected %s, got %s) on transcript %s" %
                (expected_end, found_end, name)
            )
        else:
            assert_equal(
                found_start, expected_start,
                "Unequal CDS start (expected %s, got %s) on transcript %s" %
                (expected_start, found_start, name)
            )
            assert_equal(
                found_end, expected_end, "Unequal CDS stop (expected %s, got %s) on transcript %s" %
                (expected_end, found_end, name)
            )

    assert_equal(
        len(transcripts), n * len(dtmp), "Not all transcripts found in input. Expected %s. Got %s."
        % (n * len(dtmp), len(transcripts))
    )


@attr(test="unit")
def test_get_transcript_from_args_add_three():
    for fmt in ("BED", "BigBed", "GTF2", "GFF3"):
        yield check_get_transcript_from_args_add_three, annotation_file_parser, fmt, False

    for fmt in ("BED", "BigBed", "GTF2", "GFF3"):
        yield check_get_transcript_from_args_add_three, annotation_file_parser, fmt, True


@attr(test="unit")
def test_get_transcript_from_args_add_three_multiple_files():
    for fmt in ("BED", "GTF2", "GFF3"):
        yield check_get_transcript_from_args_add_three, annotation_file_parser, fmt, False, 3


@attr(test="unit")
def test_get_transcript_from_args_multiple_bigbed_raises_error():
    files = " ".join([MINI["bigbed_file"]] * 2)
    argstr = "--annotation_format BigBed --annotation_files %s " % (files)
    parser = get_annotation_file_parser()
    args = parser.parse_args(shlex.split(argstr))
    assert_raises(SystemExit, get_transcripts_from_args, args)


@attr(test="unit")
def check_get_transcript_from_args_add_three_tabix(fmt):
    # check endpoints of transcript and CDS
    # GTF2 and BED format
    bed_ref = REF_FILES["100transcripts_bed"]
    cds = {}
    tx = {}
    with open(bed_ref) as fh:
        for line in fh:
            items = line.strip("\n").split("\t")
            name = items[3]

            tx_start = int(items[1])
            tx_end = int(items[2])
            tx[name] = (tx_start, tx_end)

            cds_start = int(items[6])
            cds_end = int(items[7])
            if cds_start < cds_end:
                cds[name] = (cds_start, cds_end)
            else:
                cds[name] = (None, None)

    ref_files = {
        "BED": REF_FILES["100transcripts_bed_tabix"],
        "GTF2": REF_FILES["100transcripts_gtf_tabix"]
    }
    parser = get_annotation_file_parser()
    argstr = "--annotation_format %s --annotation_files %s --tabix" % (fmt, ref_files[fmt])
    args = parser.parse_args(shlex.split(argstr))
    transcripts = list(get_transcripts_from_args(args))

    for my_tx in transcripts:
        name = my_tx.get_name()
        expected_cds_start, expected_cds_end = cds[name]

        found_cds_start = my_tx.cds_genome_start
        found_cds_end = my_tx.cds_genome_end

        found_tx_start = my_tx.spanning_segment.start
        found_tx_end = my_tx.spanning_segment.end

        expected_tx_start, expected_tx_end = tx[name]
        if found_cds_start is None:
            assert_true(found_cds_end is None)
            assert_true(
                expected_cds_start is None,
                "Unequal CDS start (expected %s, got %s) on transcript %s" %
                (expected_cds_start, found_cds_start, name)
            )
            assert_true(
                expected_cds_end is None, "Unequal CDS stop (expected %s, got %s) on transcript %s"
                % (expected_cds_end, found_cds_end, name)
            )
        else:
            assert_equal(
                found_cds_start, expected_cds_start,
                "Unequal CDS start (expected %s, got %s) on transcript %s" %
                (expected_cds_start, found_cds_start, name)
            )
            assert_equal(
                found_cds_end, expected_cds_end,
                "Unequal CDS stop (expected %s, got %s) on transcript %s" %
                (expected_cds_end, found_cds_end, name)
            )

        assert_equal(
            found_tx_start, expected_tx_start,
            "Unequal transcript start (expected %s, got %s) on transcript %s" %
            (expected_tx_start, found_tx_start, name)
        )
        assert_equal(
            found_tx_end, expected_tx_end,
            "Unequal transcript end (expected %s, got %s) on transcript %s" %
            (expected_tx_end, found_tx_end, name)
        )

        assert_equal(
            len(transcripts), len(cds), "Not all transcripts found in input. Expected %s. Got %s." %
            (len(cds), len(transcripts))
        )


@attr(test="unit")
def test_get_transcript_from_args_add_three_tabix():
    for fmt in ("BED", "GTF2"):
        yield check_get_transcript_from_args_add_three_tabix, fmt


def check_arg_not_raises_error(name, func, args, params, kwargs):
    params = [args] + params
    func(*params, **kwargs)


def check_arg_raises_error(name, func, args, params, kwargs):
    #    msg = "%s did not raise error with args '%s', params '%s' and kwargs '%s'" % (name,args,params,kwargs)
    params = [args] + params
    assert_raises(SystemExit, func, *params, **kwargs)


@attr(test="unit")
def test_annotation_not_sorted_raises_error_if_required():
    for func in get_transcripts_from_args, get_segmentchains_from_args:
        for fmt in ("BED", "GTF2", "GFF3"):
            argstr = "--annotation_format %s --annotation_files %s" % (
                fmt, REF_FILES["100transcripts_%s" % EXTENSIONS[fmt]]
            )
            args = annotation_file_parser.parse_args(shlex.split(argstr))
            name = "require_sort_but_not_sorted_%s" % fmt
            yield check_arg_raises_error, name, func, args, [], {"require_sort": True}


@attr(test="unit")
def test_annotation_sorted_raises_error_if_required():
    for func in get_transcripts_from_args, get_segmentchains_from_args:
        # no warning if sorted
        for fmt in ("BED", "GTF2", "GFF3"):
            argstr = "--sorted --annotation_format %s --annotation_files %s" % (
                fmt, REF_FILES["100transcripts_%s" % EXTENSIONS[fmt]]
            )
            args = annotation_file_parser.parse_args(shlex.split(argstr))
            name = "require_sort_and_sorted_%s" % fmt
            yield check_arg_not_raises_error, name, func, args, [], {"require_sort": True}
        # no warning if sorted
        for fmt in ("BED", "GTF2", "GFF3"):
            argstr = "--tabix --annotation_format %s --annotation_files %s" % (
                fmt, REF_FILES["100transcripts_%s" % EXTENSIONS[fmt]]
            )
            args = annotation_file_parser.parse_args(shlex.split(argstr))
            name = "require_sort_and_tabix_%s" % fmt
            yield check_arg_not_raises_error, name, func, args, [], {"require_sort": True}


@attr(test="unit")
def test_annotation_not_sorted_not_raises_error_if_not_required():
    for func in get_transcripts_from_args, get_segmentchains_from_args:
        for fmt in ("BED", "GTF2", "GFF3"):
            argstr = "--annotation_format %s --annotation_files %s" % (
                fmt, REF_FILES["100transcripts_%s" % EXTENSIONS[fmt]]
            )
            args = annotation_file_parser.parse_args(shlex.split(argstr))
            name = "not_require_sort_and_not_sorted_%s" % fmt
            yield check_arg_not_raises_error, name, func, args, [], {"require_sort": False}


#=============================================================================
# INDEX: tests for genome hash parsing
#=============================================================================


@attr(test="unit")
def test_mask_genome_hash_is_empty_if_no_file():
    argstr = "--mask_annotation_format BED"
    args = mask_file_parser.parse_args(shlex.split(argstr))
    mask_hash = get_genome_hash_from_mask_args(args)
    assert_true(isinstance(mask_hash, GenomeHash))
    assert_equal(len(mask_hash.feature_dict), 0)


def check_mask_genome_hash(input_format, num_features):
    """Checks that a |GenomeHash| is correctly parsed from arguments,
    by making sure that it is of the right subclass, and that it
    contains the expected number of features

    Parameters
    ----------
    input_format : str
        Format of input file ("BED","BigBed", or "PSL")

    num_features : int
        Number of features in input file
    """
    argstr = "--mask_annotation_format %s --mask_annotation_files %s" % (
        input_format, MINI["%s_file" % input_format.lower()]
    )
    args = mask_file_parser.parse_args(shlex.split(argstr))
    mask_hash = get_genome_hash_from_mask_args(args)

    # check correct return type and make sure features are populated
    if input_format == "BigBed":
        assert_true(isinstance(mask_hash, BigBedGenomeHash))
        assert_equal(mask_hash.bigbedreaders[0].num_records, num_features)
    else:
        assert_true(isinstance(mask_hash, GenomeHash))
        assert_equal(len(mask_hash.feature_dict), num_features)


@attr(test="unit")
def test_mask_genome_hash():
    # psl is missing 2 features corresponding to introns in other files
    for input_format, num_features in [("BED", 12), ("BigBed", 12), ("PSL", 10)]:
        yield check_mask_genome_hash, input_format, num_features


#=============================================================================
# INDEX: tests for sequence file parsing
#=============================================================================


def test_sequence_file_parser_opens_formats():
    for fmt, index in [("fasta", True), ("fasta", False), ("twobit", True)]:

        yield check_sequence_file_parser_opens_format, fmt, index


def check_sequence_file_parser_opens_format(fmt, index):
    argstr = "--sequence_file %s --sequence_format %s" % (REF_FILES["yeast_%s" % fmt], fmt)
    args = sequence_file_parser.parse_args(shlex.split(argstr))
    seqdict = get_seqdict_from_args(args, index)
    ref_chroms = sorted(TwoBitFile(REF_FILES["yeast_twobit"]).keys())
    msg = "Sequence file parser failed to open format '%s'." % fmt
    assert_list_equal(ref_chroms, sorted(seqdict.keys()), msg)


#=============================================================================
# INDEX: tests for PrefixNamespaceWrapper
#=============================================================================


@attr(test="unit")
class TestPrefixNamespaceWrapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # tuples of:
        #     argument name, argument type, default value
        cls.parser_options = [
            ("foo", int, 3),
            ("bar", str, "barbarbar"),
            ("baz", float, 4.5),
        ]

    def _get_parser(self, prefix=""):
        parser = argparse.ArgumentParser()
        for name, type_, val in self.parser_options:
            parser.add_argument("--%s%s" % (prefix, name), type=type_, default=val)

        return parser

    def _get_args(self, parser, fake_argv=[]):
        return parser.parse_args(fake_argv)

    def test_no_prefix(self):
        self._prefix_helper("")

    def test_prefix(self):
        self._prefix_helper("foofix_")

    def _prefix_helper(self, prefix):
        args = self._get_args(self._get_parser(prefix=prefix))
        wrapped = PrefixNamespaceWrapper(args, prefix)

        c = 0
        for name, type_, val in self.parser_options:
            c += 1
            wrapped_val = getattr(wrapped, name)
            raw_val = getattr(args, "%s%s" % (prefix, name))
            self.assertEqual(wrapped_val, val)
            self.assertEqual(raw_val, val)
            self.assertTrue(isinstance(wrapped_val, type_))

        self.assertEqual(c, len(self.parser_options))



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.crossmap`"""
import tempfile
import os
import subprocess

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import (
    RPATH,
    REF_FILES,
    COUNT_OPTIONS,
    ANNOTATION_OPTIONS,
    MASK_OPTIONS,
)

from plastid.bin.crossmap import main
from plastid.util.services.decorators import catch_stderr

#===============================================================================
# INDEX: global constants used by tests
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.crossmap",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="crossmap"),
}
"""Dictionary of global variables used in this module"""

# path to bowtie index, which will be built
_ebwt = os.path.join(test_info["temp_file_path"], "yeast_ebwt")

# equality options used by tests
_eq_ops = "--no_header --sort_keys 3"

# command-line string used to execute crossmap script
# substitute: ebwt, outbase, sequence file, offset, mismatches
_exec_str = "%s %s --sequence_file %s --offset %s --mismatches %s -k 26 -p %s --save_kmers"

# basename for output files for a test run
# substitute: offset, mismatches
_outbase_str = "yeast_crossmap_o%s"

# kmer file basename, generated by first test and used by remaining tests
_kmers = os.path.join(test_info["temp_file_path"], ((_outbase_str % 0) + "_26_0"))

#===============================================================================
# INDEX: definition of tests
#        The first we specify explicitly,
#        the remaining we generate programmatically
#
# There are four tests total:
# 
#    1. 0 0 mismatches, 0 offset, k=26, single process, and save k-mers
#    2. 0 mismatches, 12 offset, k=26, single process, use pre-existing k-mers
#    3. 2 mismatches, 12 offset, k=26, single process, use pre-existing k-mers
#    4. 0 mismatches, 12 offset, k=26, four processes, use pre-existing k-mers
#===============================================================================

# first test generates kmer file
offset = mismatches = 0
outbase = _outbase_str % offset
crossmap_tests = [
    (
        _exec_str % (
            _ebwt, os.path.join(test_info["temp_file_path"], outbase), REF_FILES["yeast_fasta"],
            offset, mismatches, 1
        ), [REF_FILES[outbase + "_26_%s" % mismatches]],
        [os.path.join(test_info["temp_file_path"], outbase + ("_26_%s_crossmap.bed" % mismatches))],
        [_eq_ops]
    ),
]
"""Tests to execute, as 4-tuples of:

    - command-line arguments to pass to :py:mod:`plastid.bin.crossmap`
    - a list of reference files corresponding to reference output
    - a list of output files made by running :py:mod:`plastid.bin.crossmap`
    - a list of equality options under which comparisons between the output
      and reference files should be evaluated
"""

# mismatch, offset pairs for remaining tests
_tuples = [(0, 12), (2, 12)]

# generate remaining tests programmatically using kmer file from first run
processes = 1
for mismatches, offset in _tuples:
    outbase = _outbase_str % offset
    test_str = (_exec_str + " --have_kmers") % (
        os.path.join(test_info["temp_file_path"], _ebwt),
        os.path.join(test_info["temp_file_path"], outbase),
        os.path.join(test_info["temp_file_path"], _kmers), offset, mismatches, processes
    )
    ref_file = REF_FILES[outbase + "_26_%s" % mismatches]
    test_file = os.path.join(
        test_info["temp_file_path"], outbase + ("_26_%s_crossmap.bed" % mismatches)
    )
    crossmap_tests.append([test_str, [ref_file], [test_file], [_eq_ops]])

# multiprocess test
mismatches = 0
processes = 4
test_str = (_exec_str + " --have_kmers") % (
    os.path.join(test_info["temp_file_path"], _ebwt),
    os.path.join(test_info["temp_file_path"], outbase + "_p4"),
    os.path.join(test_info["temp_file_path"], _kmers), offset, mismatches, processes
)
ref_file = REF_FILES[outbase + "_26_%s" % mismatches]
test_file = os.path.join(
    test_info["temp_file_path"], outbase + "_p4" + ("_26_%s_crossmap.bed" % mismatches)
)
crossmap_tests.append([test_str, [ref_file], [test_file], [_eq_ops]])

#===============================================================================
# INDEX: Helper functions to run tests
#===============================================================================


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for plastid.bin.crossmap"""
    for x in execute_helper(test_info, crossmap_tests):
        yield x


def setUpModule():
    """Create bowtie index before running tests"""
    args = ["bowtie-build", REF_FILES["yeast_fasta"], _ebwt]
    with open(os.devnull, "a") as null:
        try:
            subprocess.check_call(args, stdout=null, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            print("Could not build bowtie index for crossmap test.")
            raise e



#!/usr/bin/env python
"""
"""
import numpy

from nose.plugins.attrib import attr
from nose.tools import assert_equal, assert_true, assert_set_equal, assert_less_equal
from numpy import nan

from plastid.util.io.filters import SkipBlankReader
from plastid.genomics.roitools import SegmentChain, GenomicSegment
from plastid.genomics.genome_hash import GenomeHash
from plastid.readers.gff import GFF3_TranscriptAssembler
from plastid.bin.metagene import window_landmark, \
                                     window_cds_start, \
                                     window_cds_stop, \
                                     group_regions_make_windows
from plastid.util.services.mini2to3 import cStringIO

_FLANKS = [
    (0,100),
    (50,100),
    (100,50),
    (100,0),
]
"""Regions upstream and downstream from landmarks to request"""


def check_equality(roi1, roi2, err_str=""):
    """Make sure two |SegmentChain| are equal by testing equality of position sets, chromosomes, and strands"""
    assert_set_equal(
        roi1.get_position_set(), roi2.get_position_set(),
        "%s ROIs unequal: unequal position sets %s vs %s" % (err_str, str(roi1), str(roi2))
    )
    if roi1.length > 0:
        assert_equal(
            roi1.spanning_segment.strand, roi2.spanning_segment.strand,
            "%s ROI strand mismatch: %s vs %s" %
            (err_str, roi1.spanning_segment.strand, roi2.spanning_segment.strand)
        )
        assert_equal(
            roi1.spanning_segment.chrom, roi2.spanning_segment.chrom,
            "%s ROI chromosome mismatch: %s vs %s" %
            (err_str, roi1.spanning_segment.chrom, roi2.spanning_segment.chrom)
        )


def check_window(
        tx_ivc,
        known_roi,
        known_offset,
        known_ref_point,
        flank_up,
        flank_down,
        test_method,
        test_name,
        ref_delta=0
):
    """Helper function to test output of window landmark functions
    
    Parameters
    ----------
    tx_ivc : |SegmentChain|
        Test Transcript from which window will be derived
    
    known_roi : |SegmentChain|
        Reference output for ROI
    
    known_offset : int
        Known offset to start of ROI
    
    known_ref_point : (str,int,str) or numpy.nan
        Known offset to landmark in ROI as ("chromosome_name",position,"strand")
    
    flank_up : int
        Flank upstream of landmark to include in ROI
    
    flank_down : int
        Flank downstream of landmark to include in ROI
    
    test_method : function
        Function to test (e.g. :py:func:`window_cds_start`, py:func:`window_cds_stop`)
    
    test_name : str
        Name of test (for generating rich error output)
    
    ref_delta : int, optional
        Distance from reference landmark at which to center windows
    """
    err_str = (
        "Failed %s on %s (strand: '%s', up: %s, down: %s). " %
        (test_name, str(tx_ivc), tx_ivc.spanning_segment.strand, flank_up, flank_down)
    ) + "%s unequal (%s vs %s)"
    test_roi, test_offset, test_ref_point = test_method(
        tx_ivc, flank_up, flank_down, ref_delta=ref_delta
    )
    check_equality(SegmentChain.from_str(known_roi), test_roi)

    # if no landmark
    if numpy.isnan(known_offset) or isinstance(known_ref_point,
                                               float) and numpy.isnan(known_ref_point):
        assert_true(numpy.isnan(test_offset), msg=err_str % ("offset", known_offset, test_offset))
        assert_true(
            numpy.isnan(test_ref_point),
            msg=err_str % ("ref_point", known_ref_point, test_ref_point)
        )
    # if landmark
    else:
        assert_equal(known_offset, test_offset, msg=err_str % ("offset", known_offset, test_offset))
        assert_equal(
            known_ref_point,
            test_ref_point,
            msg=err_str % ("ref_point", known_ref_point, test_ref_point)
        )


@attr(test="unit")
def test_window_cds_start():
    # test case, 1 transcript with CDS at first nucleotide
    # test case, 1 transcript with CDS internal
    # test case, 1 transcript with no CDS

    # for all, with and without changed upstream, downstream flanks
    for flank_up, flank_down in _FLANKS:
        for txid in _CDS_START_QUERIES:
            tx_ivc = _TRANSCRIPTS[txid]
            known_vals = _CDS_START_RESULTS["%s_%s_%s" % (tx_ivc.get_name(), flank_up, flank_down)]
            ltmp = [check_window, tx_ivc]
            ltmp.extend(known_vals)
            ltmp.append(flank_up)
            ltmp.append(flank_down)
            ltmp.append(window_cds_start)
            ltmp.append("window_cds_start")
            yield tuple(ltmp)


@attr(test="unit")
def test_window_cds_stop():
    # test case, 1 transcript with CDS stop at final nucleotide
    # test case, 1 transcript with CDS stop internal
    # test case, 1 transcript with no CDS

    # for all, with and without changed upstream, downstream flanks
    for flank_up, flank_down in _FLANKS:
        for txid in _CDS_STOP_QUERIES:
            tx_ivc = _TRANSCRIPTS[txid]
            known_vals = _CDS_STOP_RESULTS["%s_%s_%s" % (tx_ivc.get_name(), flank_up, flank_down)]
            ltmp = [check_window, tx_ivc]
            ltmp.extend(known_vals)
            ltmp.append(flank_up)
            ltmp.append(flank_down)
            ltmp.append(window_cds_stop)
            ltmp.append("window_cds_stop")
            yield tuple(ltmp)


@attr(test="unit")
def test_window_cds_stop_with_ref_delta():
    # test case, 1 transcript with CDS stop at final nucleotide
    # test case, 1 transcript with CDS stop internal
    # test case, 1 transcript with no CDS

    # for all, with and without changed upstream, downstream flanks
    for flank_up, flank_down in _FLANKS:
        for txid in _CDS_STOP_QUERIES:
            tx_ivc = _TRANSCRIPTS[txid]
            known_vals = _CDS_STOP_WITH_DELTA_RESULTS["%s_%s_%s" %
                                                      (tx_ivc.get_name(), flank_up, flank_down)]
            ltmp = [check_window, tx_ivc]
            ltmp.extend(known_vals)
            ltmp.append(flank_up)
            ltmp.append(flank_down)
            ltmp.append(window_cds_stop)
            ltmp.append("window_cds_stop")
            ltmp.append(3)
            yield tuple(ltmp)


def check_window_landmark(ivc, landmark, flank_up=50, flank_down=100):
    test_roi, test_offset, (test_chrom,test_pos,test_strand) = window_landmark(ivc,flank_up,flank_down,landmark)

    # make sure position of interest matches what we gave window_landmark
    ref_chrom, ref_pos, ref_strand = ivc.get_genomic_coordinate(landmark)
    assert_equal(ref_chrom, test_chrom)
    assert_equal(ref_pos, test_pos)
    assert_equal(ref_strand, test_strand)

    # make sure position is in roi
    assert_true(test_pos in test_roi.get_position_set())

    # assure test roi length + offset == flank_up + flank_down
    # which it always should, unless the landmark is very close
    # to the edge of the transcript, in which case it should be smaller
    if landmark + flank_down <= test_roi.length:
        assert_equal(test_offset + test_roi.length, flank_up + flank_down)
    else:
        assert_less_equal(test_offset + test_roi.length, flank_up + flank_down)

    # test offset in roi is correct relative to offset and flank
    # only relevant for plus-strand
    roi_pos = test_roi.get_segmentchain_coordinate(test_chrom, test_pos, test_strand)
    assert_equal(roi_pos + test_offset, flank_up)


@attr(test="unit")
def test_window_landmark():
    # test cases: plus and minus-strand IVCs with splicing
    flank_up = 50
    flank_down = 100
    my_segmentchains = [
        SegmentChain(GenomicSegment("chrA", 50, 350, "+"), GenomicSegment("chrA", 500, 900, "+")),
        SegmentChain(GenomicSegment("chrA", 50, 350, "-"), GenomicSegment("chrA", 500, 900, "-")),
    ]
    for my_segmentchain in my_segmentchains:
        for landmark in range(0, 700, 50):
            yield check_window_landmark, my_segmentchain, landmark, flank_up, flank_down


def check_maximal_window(test_name, genome_hash, test_group, result_groups, flank_up, flank_down):
    """
    test_name : str
        Descriptive name of test

    genome_hash : GenomeHash
        Mask hash

    test_group : list
        List of transcript IDs, referring to transcripts in the GFF text above

    result_groups : list
        list of tuples of (region_str, aligment_offset, window_length) expected from 
        maximal spanning window output

    flank_up : int
        Bases to include upstream of landmark in window

    flank_down : int
        bases to include downstream of landmark in window
    """
    # table keys:
    #    gene_id
    #    window_size
    #    roi
    #    masked
    #    alignment_offset
    #    zero_point
    err_str = ("Failed %s (up: %s, down: %s). " %
               (test_name, flank_up, flank_down)) + "%s unequal (%s vs %s)"
    tx_ivcs = (_TRANSCRIPTS[X] for X in test_group)
    roi_table = group_regions_make_windows(
        tx_ivcs, genome_hash, flank_up, flank_down, window_cds_start
    )
    roi_table.sort_values("region", inplace=True)
    trows = [X[1] for X in roi_table.iterrows()]
    result_groups = sorted(result_groups, key=lambda x: x[0])
    REGION = 0

    c = 0

    for n, result_group in enumerate(result_groups):
        # if no landmark
        if numpy.isnan(result_group[1]) or numpy.isnan(result_group[2]):
            c += 1  # increment counter for input that will have no output

        # if landmark
        else:
            check_equality(
                SegmentChain.from_str(result_group[0]),
                SegmentChain.from_str(trows[n - c]["region"]), test_name
            )
            assert_equal(
                result_group[1],
                trows[n - c]["alignment_offset"],
                msg=err_str % ("offset", result_group[1], trows[n - c]["alignment_offset"])
            )
            assert_equal(
                result_group[2],
                trows[n - c]["zero_point"],
                msg=err_str % ("ref_point", result_group[1], trows[n - c]["zero_point"])
            )
            if len(result_group) == 4:
                assert_equal(
                    result_group[3],
                    trows[n - c]["masked"],
                    msg=err_str % ("mask", result_group[3], trows[n - c]["masked"])
                )

    assert_equal(n + 1 - c, len(roi_table))


@attr(test="unit")
def test_group_regions_make_windows_finds_maximal_window():
    # test case, 3 transcripts same start
    # test case, 3 transcripts, different starts
    # test case, 3 transcripts, 2 share start, 1 different
    # test case including ncRNAs
    # for all, with and without changed upstream, downstream flanks
    empty_hash = GenomeHash([])
    for flank_up, flank_down in _FLANKS:
        for test_name, test_group in _DO_GENERATE_MAX_WINDOW.items():
            result_group = _DO_GENERATE_MAX_WINDOW_RESULTS["%s_%s_%s" %
                                                           (test_name, flank_up, flank_down)]
            yield check_maximal_window, test_name, empty_hash, test_group, [result_group], flank_up, flank_down


@attr(test="unit")
def test_group_regions_make_windows_multiple_genes():
    empty_hash = GenomeHash([])
    flank_up = 50
    flank_down = 100
    for test_name, test_group in _DO_GENERATE_MULTI_GENE.items():
        result_groups = _DO_GENERATE_MULTI_GENE_RESULTS["%s_%s_%s" %
                                                        (test_name, flank_up, flank_down)]
        yield check_maximal_window, test_name, empty_hash, test_group, result_groups, flank_up, flank_down


@attr(test="unit")
def test_group_regions_make_windows_finds_masks():
    crossmap = GenomeHash(_MASKS)
    for flank_up, flank_down in _FLANKS:
        for test_name, test_group in _DO_GENERATE_MAX_WINDOW.items():
            result_group = _DO_GENERATE_MAX_WINDOW_RESULTS_MASKED["%s_%s_%s" %
                                                                  (test_name, flank_up, flank_down)]
            yield check_maximal_window, test_name, crossmap, test_group, [result_group], flank_up, flank_down


#===============================================================================
# INDEX: test data
#===============================================================================

_MASKS = [
    SegmentChain.from_str("2L:7985694-7985744(+)"),
    SegmentChain.from_str("3R:4519879-4519891(-)"),
    SegmentChain.from_str("4:50-50000(+)"),
]

_TRANSCRIPTS_GFF = """##gff-version 3
3R    FlyBase    mRNA    4517211    4523544    .    -    .    ID=FBtr0081950;Name=hb-RB;Parent=FBgn0001180;Alias=FBtr0002097,FBtr0002098,CG9786-RB,hb[+]R2.8;Dbxref=FlyBase_Annotation_IDs:CG9786-RB,REFSEQ:NM_169234;score_text=Strongly Supported;score=11
3R    FlyBase    exon    4517211    4519894    .    -    .    Name=hb:2;Parent=FBtr0081950;parent_type=mRNA
3R    FlyBase    CDS    4517600    4519876    .    -    0    Name=hb-cds;Parent=FBtr0081950;parent_type=mRNA
3R    FlyBase    exon    4523048    4523544    .    -    .    Name=hb:4;Parent=FBtr0081950;parent_type=mRNA

3R    FlyBase    mRNA    4516702    4520322    .    -    .    ID=FBtr0081951;Name=hb-RA;Parent=FBgn0001180;Alias=FBtr0002096,FBtr0002097,CG9786-RA,hb[+]R3.2;Dbxref=FlyBase_Annotation_IDs:CG9786-RA,REFSEQ:NM_169233;score_text=Strongly Supported;score=11
3R    FlyBase    exon    4516702    4519894    .    -    .    Name=hb:1;Parent=FBtr0081951;parent_type=mRNA
3R    FlyBase    CDS    4517600    4519876    .    -    0    Name=hb-cds;Parent=FBtr0081951;parent_type=mRNA
3R    FlyBase    exon    4520178    4520322    .    -    .    Name=hb:3;Parent=FBtr0081951;parent_type=mRNA

3R    FlyBase    mRNA    4516702    4520322    .    -    .    ID=FBtr0081951_alt_start;Name=hb-RA_alt_start;Parent=FBgn0001180;Alias=FBtr0002096,FBtr0002097,CG9786-RA,hb[+]R3.2;Dbxref=FlyBase_Annotation_IDs:CG9786-RA,REFSEQ:NM_169233;score_text=Strongly Supported;score=11
3R    FlyBase    exon    4516702    4519894    .    -    .    Name=hb:1;Parent=FBtr0081951_alt_start;parent_type=mRNA
3R    FlyBase    CDS     4517600    4519816    .    -    0    Name=hb-cds;Parent=FBtr0081951_alt_start;parent_type=mRNA
3R    FlyBase    exon    4520178    4520322    .    -    .    Name=hb:3;Parent=FBtr0081951_alt_start;parent_type=mRNA

3R    FlyBase    mRNA    4517211    4523544    .    -    .    ID=FBtr0081950_no_cds;Name=hb-RB_no_cds;Parent=FBgn0001180;Alias=FBtr0002097,FBtr0002098,CG9786-RB,hb[+]R2.8;Dbxref=FlyBase_Annotation_IDs:CG9786-RB,REFSEQ:NM_169234;score_text=Strongly Supported;score=11
3R    FlyBase    exon    4517211    4519894    .    -    .    Name=hb_no_cds:2;Parent=FBtr0081950_no_cds;parent_type=mRNA
3R    FlyBase    exon    4523048    4523544    .    -    .    Name=hb_no_cds:4;Parent=FBtr0081950_no_cds;parent_type=mRNA

3R    FlyBase    mRNA    4517590    4523544    .    -    .    ID=FBtr0081950_short_utr;Name=hb-RB_short_utr;Parent=FBgn0001180;Alias=FBtr0002097,FBtr0002098,CG9786-RB,hb[+]R2.8;Dbxref=FlyBase_Annotation_IDs:CG9786-RB,REFSEQ:NM_169234;score_text=Strongly Supported;score=11
3R    FlyBase    exon    4517590    4519894    .    -    .    Name=hb:2;Parent=FBtr0081950_short_utr;parent_type=mRNA
3R    FlyBase    CDS    4517600    4519876    .    -    0    Name=hb-cds;Parent=FBtr0081950_short_utr;parent_type=mRNA

3R    FlyBase    mRNA    4517600    4519876    .    -    .    ID=FBtr0081950_no_utr;Name=hb-RB_no_utr;Parent=FBgn0001180;Alias=FBtr0002097,FBtr0002098,CG9786-RB,hb[+]R2.8;Dbxref=FlyBase_Annotation_IDs:CG9786-RB,REFSEQ:NM_169234;score_text=Strongly Supported;score=11
3R    FlyBase    exon    4517600    4519876    .    -    .    Name=hb:2;Parent=FBtr0081950_no_utr;parent_type=mRNA
3R    FlyBase    CDS    4517600    4519876    .    -    0    Name=hb-cds;Parent=FBtr0081950_no_utr;parent_type=mRNA

3R    FlyBase    mRNA    4517211    4523544    .    -    .    ID=FBtr0081950_at_splice;Name=hb-RB_at_splice;Parent=FBgn0001180;Alias=FBtr0002097,FBtr0002098,CG9786-RB,hb[+]R2.8;Dbxref=FlyBase_Annotation_IDs:CG9786-RB,REFSEQ:NM_169234;score_text=Strongly Supported;score=11
3R    FlyBase    exon    4523500    4523544    .    -    .    Name=hb:3;Parent=FBtr0081950_at_splice;parent_type=mRNA
3R    FlyBase    exon    4517211    4517241    .    -    .    Name=hb:2;Parent=FBtr0081950_at_splice;parent_type=mRNA
3R    FlyBase    exon    4517600    4519876    .    -    .    Name=hb:2g;Parent=FBtr0081950_at_splice;parent_type=mRNA
3R    FlyBase    CDS    4517600    4519876    .    -    0    Name=hb-cds;Parent=FBtr0081950_at_splice;parent_type=mRNA

2L    FlyBase    mRNA    7983946    7986785    .    +    .    ID=FBtr0300185;Name=CG7231-RE;Parent=FBgn0031968;Dbxref=FlyBase_Annotation_IDs:CG7231-RE,REFSEQ:NM_164776;score_text=Moderately Supported;score=7
2L    FlyBase    CDS    7984768    7984810    .    +    0    Name=CG7231-cds;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    CDS    7985553    7985768    .    +    2    Name=CG7231-cds;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    CDS    7986139    7986705    .    +    0    Name-=CG7231-cds;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    CDS    7985834    7986084    .    +    2    Name=CG7231-cds;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    exon    7983946    7984238    .    +    .    Name=CG7231:1;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    exon    7984541    7984810    .    +    .    Name=CG7231:2;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    exon    7985553    7985768    .    +    .    Name=CG7231:6b;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    exon    7985834    7986084    .    +    .    Name=CG7231:8;Parent=FBtr0300185;parent_type=mRNA
2L    FlyBase    exon    7986139    7986785    .    +    .    Name=CG7231:9b;Parent=FBtr0300185;parent_type=mRNA

2L    FlyBase    mRNA    7985283    7986785    .    +    .    ID=FBtr0079531;Name=CG7231-RC;Parent=FBgn0031968;Dbxref=FlyBase_Annotation_IDs:CG7231-RC,REFSEQ:NM_205923;score_text=Moderately Supported;score=7
2L    FlyBase    exon    7985283    7985433    .    +    .    Name=CG7231:4;Parent=FBtr0079531;parent_type=mRNA
2L    FlyBase    CDS    7985675    7985768    .    +    0    Name=CG7231-cdsc;Parent=FBtr0079531;parent_type=mRNA
2L    FlyBase    CDS    7985834    7986075    .    +    2    Name=CG7231-cdsc;Parent=FBtr0079531;parent_type=mRNA
2L    FlyBase    CDS    7986139    7986705    .    +    0    Name=CG7231-cdsc;Parent=FBtr0079531;parent_type=mRNA
2L    FlyBase    exon    7985553    7985768    .    +    .    Name=CG7231:6c;Parent=FBtr0079531;parent_type=mRNA
2L    FlyBase    exon    7985834    7986075    .    +    .    Name=CG7231:7c;Parent=FBtr0079531;parent_type=mRNA
2L    FlyBase    exon    7986139    7986785    .    +    .    Name=CG7231:9c;Parent=FBtr0079531;parent_type=mRNA

2L    FlyBase    mRNA    7985283    7986785    .    +    .    ID=FBtr0306336;Name=CG7231-RF;Parent=FBgn0031968;Dbxref=REFSEQ:NM_001259004,FlyBase_Annotation_IDs:CG7231-RF;score_text=Moderately Supported;score=7
2L    FlyBase    exon    7985283    7985441    .    +    .    Name=CG7231:5;Parent=FBtr0306336;parent_type=mRNA
2L    FlyBase    CDS    7985675    7985768    .    +    0    Name=CG7231-cds;Parent=FBtr0306336;parent_type=mRNA
2L    FlyBase    CDS    7985834    7986075    .    +    2    Name=CG7231-cds;Parent=FBtr0306336;parent_type=mRNA
2L    FlyBase    CDS    7986139    7986705    .    +    0    Name=CG7231-cds;Parent=FBtr0306336;parent_type=mRNA
2L    FlyBase    exon    7985553    7985768    .    +    .    Name=CG7231:6;Parent=FBtr0306336;parent_type=mRNA
2L    FlyBase    exon    7985834    7986075    .    +    .    Name=CG7231:7;Parent=FBtr0306336;parent_type=mRNA
2L    FlyBase    exon    7986139    7986785    .    +    .    Name=CG7231:9;Parent=FBtr0306336;parent_type=mRNA

2L    FlyBase    mRNA    7985665    7986715    .    +    .    ID=FBtr0079531_short_utr;Name=CG7231-RC;Parent=FBgn0031968;Dbxref=FlyBase_Annotation_IDs:CG7231-RC,REFSEQ:NM_205923;score_text=Moderately Supported;score=7
2L    FlyBase    exon    7985665    7985768    .    +    .    Name=CG7231:6;Parent=FBtr0079531_short_utr;parent_type=mRNA
2L    FlyBase    exon    7985834    7986075    .    +    .    Name=CG7231:7;Parent=FBtr0079531_short_utr;parent_type=mRNA
2L    FlyBase    exon    7986139    7986715    .    +    .    Name=CG7231:9;Parent=FBtr0079531_short_utr;parent_type=mRNA
2L    FlyBase    CDS    7985675    7985768    .    +    0    Name=CG7231-cds;Parent=FBtr0079531_short_utr;parent_type=mRNA
2L    FlyBase    CDS    7985834    7986075    .    +    2    Name=CG7231-cds;Parent=FBtr0079531_short_utr;parent_type=mRNA
2L    FlyBase    CDS    7986139    7986705    .    +    0    Name=CG7231-cds;Parent=FBtr0079531_short_utr;parent_type=mRNA

2L    FlyBase    mRNA    7985283    7986785    .    +    .    ID=FBtr0079531_no_cds;Name=CG7231-RC;Parent=FBgn0031968;Dbxref=FlyBase_Annotation_IDs:CG7231-RC,REFSEQ:NM_205923;score_text=Moderately Supported;score=7
2L    FlyBase    exon    7985283    7985433    .    +    .    Name=CG7231:4;Parent=FBtr0079531_no_cds;parent_type=mRNA
2L    FlyBase    exon    7985553    7985768    .    +    .    Name=CG7231:6;Parent=FBtr0079531_no_cds;parent_type=mRNA
2L    FlyBase    exon    7985834    7986075    .    +    .    Name=CG7231:7;Parent=FBtr0079531_no_cds;parent_type=mRNA
2L    FlyBase    exon    7986139    7986785    .    +    .    Name=CG7231:9;Parent=FBtr0079531_no_cds;parent_type=mRNA

2L    FlyBase    mRNA    7985675    7986075    .    +    .    ID=FBtr0079531_no_utr;Name=CG7231-RC;Parent=FBgn0031968;Dbxref=FlyBase_Annotation_IDs:CG7231-RC,REFSEQ:NM_205923;score_text=Moderately Supported;score=7
2L    FlyBase    exon    7985675    7985768    .    +    .    Name=CG7231:6;Parent=FBtr0079531_no_utr;parent_type=mRNA
2L    FlyBase    exon    7985834    7986075    .    +    .    Name=CG7231:7;Parent=FBtr0079531_no_utr;parent_type=mRNA
2L    FlyBase    CDS    7985675    7985768    .    +    0    Name=CG7231-cds;Parent=FBtr0079531_no_utr;parent_type=mRNA
2L    FlyBase    CDS    7985834    7986075    .    +    2    Name=CG7231-cds;Parent=FBtr0079531_no_utr;parent_type=mRNA
2L    FlyBase    CDS    7986139    7986705    .    +    0    Name=CG7231-cds;Parent=FBtr0079531_no_utr;parent_type=mRNA

2L    FlyBase    mRNA    7985283    7986785    .    +    .    ID=FBtr0079531_at_splice;Name=CG7231-RC;Parent=FBgn0031968;Dbxref=FlyBase_Annotation_IDs:CG7231-RC,REFSEQ:NM_205923;score_text=Moderately Supported;score=7
2L    FlyBase    exon    7985283    7985433    .    +    .    Name=CG7231:4;Parent=FBtr0079531_at_splice;parent_type=mRNA
2L    FlyBase    exon    7985675    7985768    .    +    .    Name=CG7231:6;Parent=FBtr0079531_at_splice;parent_type=mRNA
2L    FlyBase    exon    7985834    7986075    .    +    .    Name=CG7231:7;Parent=FBtr0079531_at_splice;parent_type=mRNA
2L    FlyBase    exon    7986139    7986705    .    +    .    Name=CG7231:9;Parent=FBtr0079531_at_splice;parent_type=mRNA
2L    FlyBase    exon    7986775    7986785    .    +    .    Name=CG7231:9;Parent=FBtr0079531_at_splice;parent_type=mRNA
2L    FlyBase    CDS    7985675    7985768    .    +    0    Name=CG7231-cds;Parent=FBtr0079531_at_splice;parent_type=mRNA
2L    FlyBase    CDS    7985834    7986075    .    +    2    Name=CG7231-cds;Parent=FBtr0079531_at_splice;parent_type=mRNA
2L    FlyBase    CDS    7986139    7986705    .    +    0    Name=CG7231-cds;Parent=FBtr0079531_at_splice;parent_type=mRNA

2L    FlyBase    mRNA    8997641    8998544    .    +    .    ID=FBtr0079763;Name=CG18661-RA;Parent=FBgn0040964;Alias=CG18661-RB;Dbxref=FlyBase_Annotation_IDs:CG18661-RA,REFSEQ:NM_144305;score_text=Strongly Supported;score=15
2L    FlyBase    mRNA    8997641    8998544    .    +    .    ID=FBtr0303880;Name=CG18661-RB;Parent=FBgn0040964;Dbxref=FlyBase_Annotation_IDs:CG18661-RB,REFSEQ:NM_001201814;score_text=Strongly Supported;score=15
2L    FlyBase    exon    8997641    8997668    .    +    .    ID=FBgn0040964:1;Name=CG18661:1;Parent=FBtr0079763,FBtr0303880;parent_type=mRNA
2L    FlyBase    CDS    8997664    8997668    .    +    0    ID=CDS_FBgn0040964:1_866;Name=CG18661-cds;Parent=FBtr0303880;parent_type=mRNA
2L    FlyBase    CDS    8997664    8997668    .    +    0    ID=CDS_FBgn0040964:1_867;Name=CG18661-cds;Parent=FBtr0079763;parent_type=mRNA
2L    FlyBase    exon    8997723    8998544    .    +    .    ID=FBgn0040964:2;Name=CG18661:2;Parent=FBtr0079763;parent_type=mRNA
2L    FlyBase    CDS    8997723    8998386    .    +    1    ID=CDS_FBgn0040964:2_867;Name=CG18661-cds;Parent=FBtr0079763;parent_type=mRNA
2L    FlyBase    exon    8997762    8998544    .    +    .    ID=FBgn0040964:3;Name=CG18661:3;Parent=FBtr0303880;parent_type=mRNA
2L    FlyBase    CDS    8997762    8998386    .    +    1    ID=CDS_FBgn0040964:3_866;Name=CG18661-cds;Parent=FBtr0303880;parent_type=mRNA

2L    FlyBase    mRNA    9397110    9397988    .    -    .    ID=FBtr0079813;Name=CG4438-RA;Parent=FBgn0032115;
2L    FlyBase    mRNA    9397110    9397988    .    -    .    ID=FBtr0303900;Name=CG4438-RB;Parent=FBgn0032115;
2L    FlyBase    exon    9397110    9397772    .    -    .    ID=FBgn0032115:2;Name=CG4438:2;Parent=FBtr0079813;parent_type=mRNA
2L    FlyBase    exon    9397110    9397764    .    -    .    ID=FBgn0032115:1;Name=CG4438:1;Parent=FBtr0303900;parent_type=mRNA
2L    FlyBase    CDS    9397177    9397746    .    -    0    ID=CDS_FBgn0032115:1_867;Name=CG4438-cds;Parent=FBtr0303900;parent_type=mRNA
2L    FlyBase    CDS    9397177    9397746    .    -    0    ID=CDS_FBgn0032115:2_867;Name=CG4438-cds;Parent=FBtr0079813;parent_type=mRNA
2L    FlyBase    exon    9397833    9397988    .    -    .    ID=FBgn0032115:3;Name=CG4438:3;Parent=FBtr0079813,FBtr0303900;parent_type=mRNA
""".replace("    ", "\t")
"""GFF of transcripts used in these tests"""

_TRANSCRIPTS = {
    X.get_name(): X
    for X in GFF3_TranscriptAssembler(SkipBlankReader(cStringIO.StringIO(_TRANSCRIPTS_GFF)))
}
"""|Transcript| representation of transcripts used in these tests"""

_CDS_START_QUERIES = [ 
    "FBtr0081950",           # minus-strand, CDS internal, UTR longer than flank
    "FBtr0081950_short_utr", # minus-strand, UTR will be shorter than flank
    "FBtr0081950_no_cds",    # minus-strand, no CDS
    "FBtr0081950_no_utr",    # minus-strand, CDS start at end of tx
    "FBtr0081950_at_splice", # minus-strand, CDS start at splice junction   
    "FBtr0079531",                  
    "FBtr0079531_short_utr",                  
    "FBtr0079531_no_cds",                  
    "FBtr0079531_no_utr",                  
    "FBtr0079531_at_splice",    
] # yapf: disable
"""IDs of individual |Transcript| s to use in :py:func:`test_window_cds_start`"""

_CDS_START_RESULTS = {
    'FBtr0081950_0_100'           : ('3R:4519776-4519876(-)', 0, ('3R', 4519875, '-')),
    'FBtr0081950_100_0'           : ('3R:4519876-4519894^4523047-4523129(-)',0,('3R', 4519875, '-')),
    'FBtr0081950_100_50'          : ('3R:4519826-4519894^4523047-4523129(-)',0,('3R', 4519875, '-')),
    'FBtr0081950_50_100'          : ('3R:4519776-4519894^4523047-4523079(-)',0,('3R', 4519875, '-')),

    'FBtr0081950_no_cds_0_100'    : ('na', nan, nan),
    'FBtr0081950_no_cds_100_0'    : ('na', nan, nan),
    'FBtr0081950_no_cds_100_50'   : ('na', nan, nan),
    'FBtr0081950_no_cds_50_100'   : ('na', nan, nan),
 
    'FBtr0081950_no_utr_0_100'    : ('3R:4519776-4519876(-)',0,('3R', 4519875, '-')),
    'FBtr0081950_no_utr_100_0'    : ('na', 100, ('3R', 4519875, '-')),
    'FBtr0081950_no_utr_100_50'   : ('3R:4519826-4519876(-)',100,('3R', 4519875, '-')),
    'FBtr0081950_no_utr_50_100'   : ('3R:4519776-4519876(-)',50,('3R', 4519875, '-')),

    'FBtr0081950_short_utr_0_100' : ('3R:4519776-4519876(-)',0,('3R', 4519875, '-')),
    'FBtr0081950_short_utr_100_0' : ('3R:4519876-4519894(-)',82,('3R', 4519875, '-')),
    'FBtr0081950_short_utr_100_50': ('3R:4519826-4519894(-)',82,('3R', 4519875, '-')),
    'FBtr0081950_short_utr_50_100': ('3R:4519776-4519894(-)',32,('3R', 4519875, '-')),

    'FBtr0081950_at_splice_0_100': ('3R:4519776-4519876(-)',0,('3R', 4519875, '-')),
    'FBtr0081950_at_splice_100_0': ('3R:4523499-4523544(-)',55,('3R', 4519875, '-')),
    'FBtr0081950_at_splice_100_50': ('3R:4519826-4519876^4523499-4523544(-)',55,('3R', 4519875, '-')),
    'FBtr0081950_at_splice_50_100': ('3R:4519776-4519876^4523499-4523544(-)',5,('3R', 4519875, '-')),    

    'FBtr0081950_no_utr_100_0': ('na', 100, ('3R', 4519875, '-')),
    'FBtr0081950_no_utr_100_50': ('3R:4519826-4519876(-)',100,('3R', 4519875, '-')),
    'FBtr0081950_no_utr_50_100': ('3R:4519776-4519876(-)',50,('3R', 4519875, '-')),

    'FBtr0081950_short_utr_0_100': ('3R:4519776-4519876(-)',0,('3R', 4519875, '-')),
    'FBtr0081950_short_utr_100_0': ('3R:4519876-4519894(-)',82,('3R', 4519875, '-')),
    'FBtr0081950_short_utr_100_50': ('3R:4519826-4519894(-)',82,('3R', 4519875, '-')),
    'FBtr0081950_short_utr_50_100': ('3R:4519776-4519894(-)',32,('3R', 4519875, '-')),
    
    'FBtr0079531_0_100': ('2L:7985674-7985768^7985833-7985839(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_100_0': ('2L:7985574-7985674(+)', 0, ('2L', 7985674, '+')),
    'FBtr0079531_100_50': ('2L:7985574-7985724(+)', 0, ('2L', 7985674, '+')),
    'FBtr0079531_50_100': ('2L:7985624-7985768^7985833-7985839(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_at_splice_0_100': ('2L:7985674-7985768^7985833-7985839(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_at_splice_100_0': ('2L:7985333-7985433(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_at_splice_100_50': ('2L:7985333-7985433^7985674-7985724(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_at_splice_50_100': ('2L:7985383-7985433^7985674-7985768^7985833-7985839(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_no_cds_0_100': ('na', nan, nan),
    'FBtr0079531_no_cds_100_0': ('na', nan, nan),
    'FBtr0079531_no_cds_100_50': ('na', nan, nan),
    'FBtr0079531_no_cds_50_100': ('na', nan, nan),
    'FBtr0079531_no_utr_0_100': ('2L:7985674-7985768^7985833-7985839(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_no_utr_100_0': ('na', 100, ('2L', 7985674, '+')),
    'FBtr0079531_no_utr_100_50': ('2L:7985674-7985724(+)',100,('2L', 7985674, '+')),
    'FBtr0079531_no_utr_50_100': ('2L:7985674-7985768^7985833-7985839(+)',50,('2L', 7985674, '+')),
    'FBtr0079531_short_utr_0_100': ('2L:7985674-7985768^7985833-7985839(+)',0,('2L', 7985674, '+')),
    'FBtr0079531_short_utr_100_0': ('2L:7985664-7985674(+)',90,('2L', 7985674, '+')),
    'FBtr0079531_short_utr_100_50': ('2L:7985664-7985724(+)',90,('2L', 7985674, '+')),
    'FBtr0079531_short_utr_50_100': ('2L:7985664-7985768^7985833-7985839(+)',40,('2L', 7985674, '+')),     
} # yapf: disable
"""Expected results of CDS start queries"""

_CDS_STOP_QUERIES = [ 
    "FBtr0081950",           # minus-strand, CDS internal, UTR longer than flank
    "FBtr0081950_short_utr", # minus-strand, UTR will be shorter than flank
    "FBtr0081950_no_cds",    # minus-strand, no CDS
    "FBtr0081950_no_utr",    # minus-strand, CDS start at end of tx
    "FBtr0081950_at_splice", # minus-strand, CDS start at splice junction   
    "FBtr0079531",                  
    "FBtr0079531_short_utr",                  
    "FBtr0079531_no_cds",                  
    "FBtr0079531_no_utr",                  
    "FBtr0079531_at_splice",    
] # yapf: disable
"""IDs of individual |Transcript| s to use in :py:func:`test_window_cds_stop`"""

_CDS_STOP_RESULTS = {
    'FBtr0079531_0_100': ('2L:7986702-7986785(+)', 0, ('2L', 7986702, '+')),
    'FBtr0079531_100_0': ('2L:7986602-7986702(+)', 0, ('2L', 7986702, '+')),
    'FBtr0079531_100_50': ('2L:7986602-7986752(+)', 0, ('2L', 7986702, '+')),
    'FBtr0079531_50_100': ('2L:7986652-7986785(+)', 0, ('2L', 7986702, '+')),
    'FBtr0079531_at_splice_0_100': ('2L:7986702-7986705^7986774-7986785(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_at_splice_100_0': ('2L:7986602-7986702(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_at_splice_100_50': ('2L:7986602-7986705^7986774-7986785(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_at_splice_50_100': ('2L:7986652-7986705^7986774-7986785(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_no_cds_0_100': ('na', nan, nan),
    'FBtr0079531_no_cds_100_0': ('na', nan, nan),
    'FBtr0079531_no_cds_100_50': ('na', nan, nan),
    'FBtr0079531_no_cds_50_100': ('na', nan, nan),
    'FBtr0079531_no_utr_0_100': ('2L:7986702-7986705(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_no_utr_100_0': ('2L:7986602-7986702(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_no_utr_100_50': ('2L:7986602-7986705(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_no_utr_50_100': ('2L:7986652-7986705(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_short_utr_0_100': ('2L:7986702-7986715(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_short_utr_100_0': ('2L:7986602-7986702(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_short_utr_100_50': ('2L:7986602-7986715(+)',0,('2L', 7986702, '+')),
    'FBtr0079531_short_utr_50_100': ('2L:7986652-7986715(+)',0,('2L', 7986702, '+')),
    'FBtr0081950_0_100': ('3R:4517502-4517602(-)', 0, ('3R', 4517601, '-')),
    'FBtr0081950_100_0': ('3R:4517602-4517702(-)', 0, ('3R', 4517601, '-')),
    'FBtr0081950_100_50': ('3R:4517552-4517702(-)', 0, ('3R', 4517601, '-')),
    'FBtr0081950_50_100': ('3R:4517502-4517652(-)', 0, ('3R', 4517601, '-')),
    'FBtr0081950_at_splice_0_100': ('3R:4517210-4517241^4517599-4517602(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_at_splice_100_0': ('3R:4517602-4517702(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_at_splice_100_50': ('3R:4517210-4517241^4517599-4517702(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_at_splice_50_100': ('3R:4517210-4517241^4517599-4517652(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_no_cds_0_100': ('na', nan, nan),
    'FBtr0081950_no_cds_100_0': ('na', nan, nan),
    'FBtr0081950_no_cds_100_50': ('na', nan, nan),
    'FBtr0081950_no_cds_50_100': ('na', nan, nan),
    'FBtr0081950_no_utr_0_100': ('3R:4517599-4517602(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_no_utr_100_0': ('3R:4517602-4517702(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_no_utr_100_50': ('3R:4517599-4517702(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_no_utr_50_100': ('3R:4517599-4517652(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_short_utr_0_100': ('3R:4517589-4517602(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_short_utr_100_0': ('3R:4517602-4517702(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_short_utr_100_50': ('3R:4517589-4517702(-)',0,('3R', 4517601, '-')),
    'FBtr0081950_short_utr_50_100': ('3R:4517589-4517652(-)',0,('3R', 4517601, '-'))             
} # yapf: disable
"""Expected results of queries"""

_CDS_STOP_WITH_DELTA_RESULTS = {
    'FBtr0079531_0_100': ('2L:7986705-7986785(+)', 0, ('2L', 7986705, '+')),
    'FBtr0079531_100_0': ('2L:7986605-7986705(+)', 0, ('2L', 7986705, '+')),
    'FBtr0079531_100_50': ('2L:7986605-7986755(+)', 0, ('2L', 7986705, '+')),
    'FBtr0079531_50_100': ('2L:7986655-7986785(+)', 0, ('2L', 7986705, '+')),
    'FBtr0079531_at_splice_0_100': ('2L:7986774-7986785(+)',0,('2L', 7986774, '+')),
    'FBtr0079531_at_splice_100_0': ('2L:7986605-7986705(+)',0,('2L', 7986774, '+')),
    'FBtr0079531_at_splice_100_50': ('2L:7986605-7986705^7986774-7986785(+)',0,('2L', 7986774, '+')),
    'FBtr0079531_at_splice_50_100': ('2L:7986655-7986705^7986774-7986785(+)',0,('2L', 7986774, '+')),
    'FBtr0079531_no_cds_0_100': ('na', nan, nan),
    'FBtr0079531_no_cds_100_0': ('na', nan, nan),
    'FBtr0079531_no_cds_100_50': ('na', nan, nan),
    'FBtr0079531_no_cds_50_100': ('na', nan, nan),
    'FBtr0079531_no_utr_0_100': ('na', 0, ('2L', 7986705, '+')),
    'FBtr0079531_no_utr_100_0': ('2L:7986605-7986705(+)',0,('2L', 7986705, '+')),
    'FBtr0079531_no_utr_100_50': ('2L:7986605-7986705(+)',0,('2L', 7986705, '+')),
    'FBtr0079531_no_utr_50_100': ('2L:7986655-7986705(+)',0,('2L', 7986705, '+')),
    'FBtr0079531_short_utr_0_100': ('2L:7986705-7986715(+)',0,('2L', 7986705, '+')),
    'FBtr0079531_short_utr_100_0': ('2L:7986605-7986705(+)',0,('2L', 7986705, '+')),
    'FBtr0079531_short_utr_100_50': ('2L:7986605-7986715(+)',0,('2L', 7986705, '+')),
    'FBtr0079531_short_utr_50_100': ('2L:7986655-7986715(+)',0,('2L', 7986705, '+')),
    'FBtr0081950_0_100': ('3R:4517499-4517599(-)', 0, ('3R', 4517598, '-')),
    'FBtr0081950_100_0': ('3R:4517599-4517699(-)', 0, ('3R', 4517598, '-')),
    'FBtr0081950_100_50': ('3R:4517549-4517699(-)', 0, ('3R', 4517598, '-')),
    'FBtr0081950_50_100': ('3R:4517499-4517649(-)', 0, ('3R', 4517598, '-')),
    'FBtr0081950_at_splice_0_100': ('3R:4517210-4517241(-)',0,('3R', 4517240, '-')),
    'FBtr0081950_at_splice_100_0': ('3R:4517599-4517699(-)',0,('3R', 4517240, '-')),
    'FBtr0081950_at_splice_100_50': ('3R:4517210-4517241^4517599-4517699(-)',0,('3R', 4517240, '-')),
    'FBtr0081950_at_splice_50_100': ('3R:4517210-4517241^4517599-4517649(-)',0,('3R', 4517240, '-')),
    'FBtr0081950_no_cds_0_100': ('na', nan, nan),
    'FBtr0081950_no_cds_100_0': ('na', nan, nan),
    'FBtr0081950_no_cds_100_50': ('na', nan, nan),
    'FBtr0081950_no_cds_50_100': ('na', nan, nan),
    'FBtr0081950_no_utr_0_100': ('na', 0, ('3R', 4517598, '-')),
    'FBtr0081950_no_utr_100_0': ('3R:4517599-4517699(-)',0,('3R', 4517598, '-')),
    'FBtr0081950_no_utr_100_50': ('3R:4517599-4517699(-)',0,('3R', 4517598, '-')),
    'FBtr0081950_no_utr_50_100': ('3R:4517599-4517649(-)',0,('3R', 4517598, '-')),
    'FBtr0081950_short_utr_0_100': ('3R:4517589-4517599(-)',0,('3R', 4517598, '-')),
    'FBtr0081950_short_utr_100_0': ('3R:4517599-4517699(-)',0,('3R', 4517598, '-')),
    'FBtr0081950_short_utr_100_50': ('3R:4517589-4517699(-)',0,('3R', 4517598, '-')),
    'FBtr0081950_short_utr_50_100': ('3R:4517589-4517649(-)',0,('3R', 4517598, '-'))
} # yapf: disable

_DO_GENERATE_MAX_WINDOW = {
    "3_same_start_plus"      : ["FBtr0306336","FBtr0079531","FBtr0079531_short_utr"],
    "3_same_start_plus2"     : ["FBtr0306336","FBtr0079531","FBtr0079531_no_utr"],
    "3_diff_start_plus"      : ["FBtr0306336","FBtr0079531","FBtr0079531_short_utr","FBtr0300185"],
    "3_same_plus_ncrna_plus" : ["FBtr0306336","FBtr0079531","FBtr0079531_short_utr","FBtr0079531_no_cds"],
    "3_diff_plus_ncrna_plus" : ["FBtr0306336","FBtr0079531","FBtr0079531_short_utr","FBtr0300185","FBtr0079531_no_cds"],

    "3_same_start_minus"       : ["FBtr0081950","FBtr0081951","FBtr0081950_short_utr"],
    "3_same_start_minus2"      : ["FBtr0081950","FBtr0081951","FBtr0081950_no_utr"],
    "3_diff_start_minus"       : ["FBtr0081950","FBtr0081951_alt_start","FBtr0081951"],
    "3_same_minus_ncrna_minus" : ["FBtr0081950","FBtr0081951","FBtr0081950_short_utr","FBtr0081950_no_cds"],
    "3_diff_minus_ncrna_minus" : ["FBtr0081950","FBtr0081951_alt_start","FBtr0081951","FBtr0081950_no_cds"],


    "stop_max_spanning_window_threeprime_plus"  : ["FBtr0079763","FBtr0303880"],

    "stop_max_spanning_window_threeprime_minus" : ["FBtr0079813","FBtr0303900"],

} # yapf: disable

_DO_GENERATE_MAX_WINDOW_RESULTS = {
    '3_diff_minus_ncrna_minus_0_100': ['na', nan, nan],
    '3_diff_minus_ncrna_minus_100_0': ['na', nan, nan],
    '3_diff_minus_ncrna_minus_100_50': ['na', nan, nan],
    '3_diff_minus_ncrna_minus_50_100': ['na', nan, nan],
    '3_diff_plus_ncrna_plus_0_100': ['na', nan, nan],
    '3_diff_plus_ncrna_plus_100_0': ['na', nan, nan],
    '3_diff_plus_ncrna_plus_100_50': ['na', nan, nan],
    '3_diff_plus_ncrna_plus_50_100': ['na', nan, nan],
    '3_diff_start_minus_0_100': ['na', nan, nan],
    '3_diff_start_minus_100_0': ['na', nan, nan],
    '3_diff_start_minus_100_50': ['na', nan, nan],
    '3_diff_start_minus_50_100': ['na', nan, nan],
    '3_diff_start_plus_0_100': ['na', nan, nan],
    '3_diff_start_plus_100_0': ['na', nan, nan],
    '3_diff_start_plus_100_50': ['na', nan, nan],
    '3_diff_start_plus_50_100': ['na', nan, nan],
    '3_same_minus_ncrna_minus_0_100': ['na', nan, nan],
    '3_same_minus_ncrna_minus_100_0': ['na', nan, nan],
    '3_same_minus_ncrna_minus_100_50': ['na', nan, nan],
    '3_same_minus_ncrna_minus_50_100': ['na', nan, nan],
    '3_same_plus_ncrna_plus_0_100': ['na', nan, nan],
    '3_same_plus_ncrna_plus_100_0': ['na', nan, nan],
    '3_same_plus_ncrna_plus_100_50': ['na', nan, nan],
    '3_same_plus_ncrna_plus_50_100': ['na', nan, nan],
    '3_same_start_minus2_0_100': ['3R:4519776-4519876(-)', 0, 0],
    '3_same_start_minus2_100_0': ['na', nan, nan],
    '3_same_start_minus2_100_50': ['3R:4519826-4519876(-)', 100, 100],
    '3_same_start_minus2_50_100': ['3R:4519776-4519876(-)', 50, 50],
    '3_same_start_minus_0_100': ['3R:4519776-4519876(-)', 0, 0],
    '3_same_start_minus_100_0': ['3R:4519876-4519894(-)', 82, 100],
    '3_same_start_minus_100_50': ['3R:4519826-4519894(-)', 82, 100],
    '3_same_start_minus_50_100': ['3R:4519776-4519894(-)', 32, 50],
    '3_same_start_plus2_0_100': ['2L:7985674-7985768^7985833-7985839(+)', 0, 0],
    '3_same_start_plus2_100_0': ['na', nan, nan],
    '3_same_start_plus2_100_50': ['2L:7985674-7985724(+)', 100, 100],
    '3_same_start_plus2_50_100': ['2L:7985674-7985768^7985833-7985839(+)', 50, 50],
    '3_same_start_plus_0_100': ['2L:7985674-7985768^7985833-7985839(+)', 0, 0],
    '3_same_start_plus_100_0': ['2L:7985664-7985674(+)', 90, 100],
    '3_same_start_plus_100_50': ['2L:7985664-7985724(+)', 90, 100],
    '3_same_start_plus_50_100': ['2L:7985664-7985768^7985833-7985839(+)', 40, 50],
    'stop_max_spanning_window_threeprime_minus_0_100': ['2L:9397646-9397746(-)', 0, 0],
    'stop_max_spanning_window_threeprime_minus_100_0': ['2L:9397746-9397764(-)', 0, 100],
    'stop_max_spanning_window_threeprime_minus_100_50': ['2L:9397696-9397764(-)', 82, 100],
    'stop_max_spanning_window_threeprime_minus_50_100': ['2L:9397646-9397764(-)', 32, 50],
    'stop_max_spanning_window_threeprime_plus_0_100': ['2L:8997663-8997668(+)', 0, 0],
    'stop_max_spanning_window_threeprime_plus_100_0': ['2L:8997640-8997663(+)', 77, 100],
    'stop_max_spanning_window_threeprime_plus_100_50': ['2L:8997640-8997668(+)', 77, 100],
    'stop_max_spanning_window_threeprime_plus_50_100': ['2L:8997640-8997668(+)', 27, 50],
} # yapf: disable

_DO_GENERATE_MAX_WINDOW_RESULTS_MASKED = {
    '3_diff_minus_ncrna_minus_0_100': ['na', nan, nan, 'na'],
    '3_diff_minus_ncrna_minus_100_0': ['na', nan, nan, 'na'],
    '3_diff_minus_ncrna_minus_100_50': ['na', nan, nan, 'na'],
    '3_diff_minus_ncrna_minus_50_100': ['na', nan, nan, 'na'],
    '3_diff_plus_ncrna_plus_0_100': ['na', nan, nan, 'na'],
    '3_diff_plus_ncrna_plus_100_0': ['na', nan, nan, 'na'],
    '3_diff_plus_ncrna_plus_100_50': ['na', nan, nan, 'na'],
    '3_diff_plus_ncrna_plus_50_100': ['na', nan, nan, 'na'],
    '3_diff_start_minus_0_100': ['na', nan, nan, 'na'],
    '3_diff_start_minus_100_0': ['na', nan, nan, 'na'],
    '3_diff_start_minus_100_50': ['na', nan, nan, 'na'],
    '3_diff_start_minus_50_100': ['na', nan, nan, 'na'],
    '3_diff_start_plus_0_100': ['na', nan, nan, 'na'],
    '3_diff_start_plus_100_0': ['na', nan, nan, 'na'],
    '3_diff_start_plus_100_50': ['na', nan, nan, 'na'],
    '3_diff_start_plus_50_100': ['na', nan, nan, 'na'],
    '3_same_minus_ncrna_minus_0_100': ['na', nan, nan, 'na'],
    '3_same_minus_ncrna_minus_100_0': ['na', nan, nan, 'na'],
    '3_same_minus_ncrna_minus_100_50': ['na', nan, nan, 'na'],
    '3_same_minus_ncrna_minus_50_100': ['na', nan, nan, 'na'],
    '3_same_plus_ncrna_plus_0_100': ['na', nan, nan, 'na'],
    '3_same_plus_ncrna_plus_100_0': ['na', nan, nan, 'na'],
    '3_same_plus_ncrna_plus_100_50': ['na', nan, nan, 'na'],
    '3_same_plus_ncrna_plus_50_100': ['na', nan, nan, 'na'],
    '3_same_start_minus2_0_100': ['3R:4519776-4519876(-)', 0, 0, 'na'],
    '3_same_start_minus2_100_0': ['na', nan, nan, 'na'],
    '3_same_start_minus2_100_50': ['3R:4519826-4519876(-)', 100, 100, 'na'],
    '3_same_start_minus2_50_100': ['3R:4519776-4519876(-)', 50, 50, 'na'],
    '3_same_start_minus_0_100': ['3R:4519776-4519876(-)', 0, 0, 'na'],
    '3_same_start_minus_100_0': ['3R:4519876-4519894(-)',82,100,'3R:4519879-4519891(-)'],
    '3_same_start_minus_100_50': ['3R:4519826-4519894(-)',82,100,'3R:4519879-4519891(-)'],
    '3_same_start_minus_50_100': ['3R:4519776-4519894(-)',32,50,'3R:4519879-4519891(-)'],
    '3_same_start_plus2_0_100': ['2L:7985674-7985768^7985833-7985839(+)',0,0,'2L:7985694-7985744(+)'],
    '3_same_start_plus2_100_0': ['na', nan, nan, 'na'],
    '3_same_start_plus2_100_50': ['2L:7985674-7985724(+)',100,100,'2L:7985694-7985724(+)'],
    '3_same_start_plus2_50_100': ['2L:7985674-7985768^7985833-7985839(+)',50,50,'2L:7985694-7985744(+)'],
    '3_same_start_plus_0_100': ['2L:7985674-7985768^7985833-7985839(+)',0,0,'2L:7985694-7985744(+)'],
    '3_same_start_plus_100_0': ['2L:7985664-7985674(+)', 90, 100, 'na'],
    '3_same_start_plus_100_50': ['2L:7985664-7985724(+)',90,100,'2L:7985694-7985724(+)'],
    '3_same_start_plus_50_100': ['2L:7985664-7985768^7985833-7985839(+)',40,50,'2L:7985694-7985744(+)'],
    
    
    'stop_max_spanning_window_threeprime_minus_0_100': ['2L:9397646-9397746(-)',  0,   0],
    'stop_max_spanning_window_threeprime_minus_100_0': ['2L:9397746-9397764(-)',  0,   100],
    'stop_max_spanning_window_threeprime_minus_100_50': ['2L:9397696-9397764(-)', 82,  100],
    'stop_max_spanning_window_threeprime_minus_50_100': ['2L:9397646-9397764(-)', 32,  50],
    
    'stop_max_spanning_window_threeprime_plus_0_100': ['2L:8997663-8997668(+)',   0,   0],
    'stop_max_spanning_window_threeprime_plus_100_0': ['2L:8997640-8997663(+)',   77,  100],
    'stop_max_spanning_window_threeprime_plus_100_50': ['2L:8997640-8997668(+)',  77,  100],
    'stop_max_spanning_window_threeprime_plus_50_100': ['2L:8997640-8997668(+)',  27,  50],      
} # yapf: disable

_DO_GENERATE_MULTI_GENE = {
    "3_same_start":
    _DO_GENERATE_MAX_WINDOW["3_same_start_plus"] + _DO_GENERATE_MAX_WINDOW["3_same_start_minus"],
    "3_same_start_1_ncrna":
    _DO_GENERATE_MAX_WINDOW["3_same_plus_ncrna_plus"] +
    _DO_GENERATE_MAX_WINDOW["3_same_start_minus"],
    "same_and_diff":
    _DO_GENERATE_MAX_WINDOW["3_diff_start_plus"] + _DO_GENERATE_MAX_WINDOW["3_same_start_minus"]
}

_DO_GENERATE_MULTI_GENE_RESULTS = {
    "3_same_start_50_100": [
        _DO_GENERATE_MAX_WINDOW_RESULTS["3_same_start_minus_50_100"],
        _DO_GENERATE_MAX_WINDOW_RESULTS["3_same_start_plus_50_100"],
    ],
    "3_same_start_1_ncrna_50_100": [
        _DO_GENERATE_MAX_WINDOW_RESULTS["3_same_plus_ncrna_plus_50_100"],
        _DO_GENERATE_MAX_WINDOW_RESULTS["3_same_start_minus_50_100"]
    ],
    "same_and_diff_50_100": [
        _DO_GENERATE_MAX_WINDOW_RESULTS["3_diff_start_plus_50_100"],
        _DO_GENERATE_MAX_WINDOW_RESULTS["3_same_start_minus_50_100"]
    ],
}
